#!/usr/bin/env python3

from __future__ import annotations

import argparse
import re
import sys
from typing import IO

from common import LockOpTrace, get_next_match


class LockOp:
    """
    LockOp: Represents a lock operation in the log file, including its header
    and times (held and wait times). It is tied to the LockOpTrace class.
    """

    def __init__(self, lines: list[str]):
        self.trace = LockOpTrace(lines[1:])
        self.header = lines[0]
        self.held_time = 0
        self.wait_time = 0
        self._calc_held_ms(self.header)
        self._calc_wait_ms(self.header)

    def _calc_held_ms(self, line: str) -> int:
        match = re.search(".*held: (.+?) ms.*", line)
        if match:
            self.held_time = int(match.group(1))
        else:
            raise ValueError("No held time in line: {}".format(line))

    def _calc_wait_ms(self, line: str) -> int:
        match = re.search(".*wait (.+?) ms.*", line)
        if match:
            self.wait_time = int(match.group(1))
        else:
            raise ValueError("No wait time in line: {}".format(line))

    def get_held_time(self) -> int:
        return self.held_time

    def get_wait_time(self) -> int:
        return self.wait_time

    def get_trace(self) -> LockOpTrace:
        return self.trace


class LocksGroup:
    """
    LocksGroup: Represents a group of locks, which could correspond to a test
    case or a project setup. It holds a header (name) and a list of Lock
    objects.
    """

    LOCK_PREFIX = "### "

    locks: list[LockOp]

    def __init__(self, lines: list[str]):
        self.lines = lines
        self.header = self.lines[0]
        self.locks = []

        self._read()

    def _read(self) -> None:
        current_line = 1
        while current_line < len(self.lines):
            lock_lines = self._current_lock(current_line)
            if len(lock_lines) == 0:
                raise RuntimeError("Error parsing lock")

            self.locks.append(LockOp(lock_lines))
            current_line = current_line + len(lock_lines)

    def __str__(self) -> str:
        return "".join(self.lines)

    # This function works to detect the current test or project
    def _current_lock(self, start_line: int) -> list[str]:
        next_match = get_next_match(self.lines, start_line, self.LOCK_PREFIX)
        if next_match == -1:
            return self.lines[start_line:len(self.lines)]
        return self.lines[start_line:next_match]

    def get_name(self) -> str:
        return self.header

    def get_locks(self) -> list[LockOp]:
        return self.locks

    def get_traces(self) -> list[LockOpTrace]:
        traces = []
        for lock in self.locks:
            traces.append(lock.get_trace())
        return traces

    def get_lock_held_time(self, trace: LockOpTrace) -> int:
        for lock in self.locks:
            if lock.get_trace() == trace:
                return lock.get_held_time()

        return 0

    def get_lock_wait_time(self, trace: LockOpTrace) -> int:
        for lock in self.locks:
            if lock.get_trace() == trace:
                return lock.get_wait_time()

        return 0


class GroupTimes:
    """
    GroupTimes: A tuple-like class that associates the group name with the
    held and wait times for each lock trace.
    """

    def __init__(self, group_name: str, held_time: int, wait_time: int):
        self.group_name = group_name
        self.held_time = held_time
        self.wait_time = wait_time

    def get_group_name(self) -> str:
        return self.group_name

    def get_held_time(self) -> int:
        return self.held_time

    def get_wait_time(self) -> int:
        return self.wait_time


class LockTraceManager:
    """
    LockTraceManager: Handles filtering and managing the lock traces
    and associated group times. It provides methods to filter the traces
    by time and to print the results in a sorted manner.
    """

    traces: dict[LockOpTrace, list[GroupTimes]]

    def __init__(self, traces: dict[LockOpTrace, list[GroupTimes]]):
        self.traces = traces

    # Filter the times for each trace
    def filter(self, held_time: int, wait_time: int) -> None:
        filtered_traces = dict[LockOpTrace, list[GroupTimes]]()
        for trace, times in self.traces.items():
            filtered_times = [
                time
                for time in times
                if time.get_held_time() >= held_time
                and time.get_wait_time() >= wait_time
            ]
            if len(filtered_times) > 0:
                filtered_traces[trace] = filtered_times

        self.traces = filtered_traces

    # Keep the traces that match with the params
    def match(self, match_names: list[str]) -> None:
        filtered_traces = dict[LockOpTrace, list[GroupTimes]]()
        for trace, times in self.traces.items():
            for match_name in match_names:
                if trace.match(match_name):
                    filtered_traces[trace] = times

        self.traces = filtered_traces

    # print the traces with their times for each test
    def print(
        self, sort_held_time: bool, sort_wait_time: bool, list_traces: bool
    ) -> None:
        if sort_held_time:
            for trace, times in self.traces.items():
                self.traces[trace] = sorted(
                    times, key=lambda x: x.get_held_time(), reverse=True
                )
        if sort_wait_time:
            for trace, times in self.traces.items():
                self.traces[trace] = sorted(
                    times, key=lambda x: x.get_held_time(), reverse=True
                )

        for trace, times in self.traces.items():
            trace.print()

            if not list_traces:
                for time in times:
                    print(
                        "{} held: {} ms, wait: {} ms".format(
                            time.get_group_name(),
                            time.get_held_time(),
                            time.get_wait_time(),
                        )
                    )
                print("")


class LocksFileReader:
    """
    LocksFileReader: Reads the lock file and parses the different test cases
    and groups. It extracts the relevant trace data and stores it in
    LocksGroup instances. It can also return a dictionary of traces and
    associated group times.
    """

    PROJECT_PREFIX = "###START: SNAPD PROJECT"
    TEST_PREFIX = "###START:"

    lines: list[str]
    groups: list[LocksGroup]

    def __init__(self, locks_file: IO[str]):
        self.lines = []
        self.groups = []

        self._read(locks_file)

    def _read(self, locks_file: IO[str]) -> None:
        self.lines = locks_file.readlines()

        current_line = 0
        if not self._is_project(self.lines[current_line]):
            print("First time expected to be the project start.")
            sys.exit(1)

        # Read the tests
        while current_line < len(self.lines):
            group_lines = self._current_group(current_line)
            if len(group_lines) == 0:
                raise RuntimeError("Error parsing test.")

            self.groups.append(LocksGroup(group_lines))
            current_line = current_line + len(group_lines)

    # Indicates if the line is the project declaration
    def _is_project(self, line: str) -> bool:
        return line.startswith(self.PROJECT_PREFIX)

    # This function works to detect the current test or project
    def _current_group(self, start_line: int) -> list[str]:
        next_match = get_next_match(self.lines, start_line, self.TEST_PREFIX)
        if next_match == -1:
            return self.lines[start_line:len(self.lines)]
        return self.lines[start_line:next_match]

    # Retrieve the test lines
    def get_test(self, test: str) -> str:
        for group in self.groups:
            if test in group.get_name():
                return str(group)

        return ""

    # Retrieve the times for each trace in the file
    # For each trace, there is a list with the times for each test where
    # the trace appears
    def get_traces_times(self) -> dict[LockOpTrace, list[GroupTimes]]:
        traces = dict[LockOpTrace, list[GroupTimes]]()
        for group in self.groups:
            group_traces = group.get_traces()
            for trace in group_traces:
                group_time = GroupTimes(
                    group.get_name(),
                    group.get_lock_held_time(trace),
                    group.get_lock_wait_time(trace),
                )
                if trace not in traces.keys():
                    traces[trace] = list[GroupTimes]()
                try:
                    traces[trace].append(group_time)
                except KeyError:
                    traces[trace] = list[GroupTimes]([group_time])

        return traces


"""
A locks file must:
- begin with "###START: SNAPD PROJECT"
- contain at least one group, indicated by the key string "###START"

It parses the lock information by:
- grouping its contents based on consecutive lines that begin with "###START"
- creating subgroups for each group by searching for the key string "###" and
  creating a subgroup for each instance found of consecutive lines between
  instances of that key string

One may filter data based on:
- matching string found in a sequence of sub-group lock information. 
  If a match is found, then only sub-groups that contain that match will be shown
- only showing held or wait times above a certain threshold
- a singular test
"""


def _make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="state-locks-filter",
        description="Reads a locks file and extracts lock information.",
    )
    parser.add_argument(
        "-f",
        "--locks-file",
        type=argparse.FileType("r"),
        required=True,
        help="Locks file",
    )
    parser.add_argument(
        "--test",
        default="",
        help="Show results for this test, when test is "
             "selected no other filters are applied.",
    )
    parser.add_argument(
        "--match",
        action="append",
        default=[],
        help="Show traces that match this specific word",
    )
    parser.add_argument(
        "--held-time", type=int, default=0,
        help="Include locks with longer held time"
    )
    parser.add_argument(
        "--wait-time", type=int, default=0,
        help="Include locks with longer wait time"
    )
    parser.add_argument(
        "--sort-held-time",
        action="store_true",
        help="Sort higher times first by held time",
    )
    parser.add_argument(
        "--sort-wait-time",
        action="store_true",
        help="Sort higher times first by wait time",
    )
    parser.add_argument(
        "--list-traces",
        action="store_true",
        help="Just print the resulting traces",
    )

    return parser


if __name__ == "__main__":
    parser = _make_parser()
    args = parser.parse_args()

    if args.sort_held_time and args.sort_wait_time:
        print("state-lock-filter: define just 1 sorting (by held/wait times)")
        sys.exit(1)

    locks_reader = LocksFileReader(args.locks_file)
    if args.test:
        print(locks_reader.get_test(args.test))
        sys.exit()

    trace_manager = LockTraceManager(locks_reader.get_traces_times())

    # Then keep traces with matches
    if args.match:
        trace_manager.match(args.match)

    # First filter by time
    if args.held_time > 0 or args.wait_time > 0:
        trace_manager.filter(args.held_time, args.wait_time)

    # And finally print the sorted results (if required)
    trace_manager.print(args.sort_held_time,
                        args.sort_wait_time,
                        args.list_traces)



import argparse
import yaml

parser = argparse.ArgumentParser()
parser.add_argument('edition_bump', type=int)
# Type of media:
# * A bootable removable media will have boot/bootx64.efi and boot/grubx64.efi.
#   It will not register boot entries in UEFI variables.
# * A bootable installed system (boot_entry) will have boot/bootx64.efi and boot/fbx64.efi as fallback.
#   It will also have <distro>/shimx64.efi and <distro>/grubx64.efi.
#   This will work with UEFI variables for boot entries.
# For more information see:
# See https://github.com/rhboot/shim/blob/66e6579dbf921152f647a0c16da1d3b2f40861ca/README.fallback
parser.add_argument('type', choices=['removable', 'boot_entry'])
parser.add_argument('gadget_yaml', type=argparse.FileType('r', encoding='utf-8'))
parser.add_argument('output', type=argparse.FileType('w', encoding='utf-8'))
args = parser.parse_args()

content = yaml.safe_load(args.gadget_yaml)
for partition in content['volumes']['pc']['structure']:
    if partition.get('role') == 'system-seed':
        partition['update']['edition'] += args.edition_bump
        if args.type == 'removable':
            partition['content'] = [
                {'source': 'grubx64.efi',
                'target': 'EFI/boot/grubx64.efi'},
                {'source': 'shim.efi.signed',
                 'target': 'EFI/boot/bootx64.efi'},
            ]
        elif args.type == 'boot_entry':
            partition['content'] = [
                {'source': 'grubx64.efi',
                 'target': 'EFI/ubuntu/grubx64.efi'},
                {'source': 'shim.efi.signed',
                 'target': 'EFI/ubuntu/shimx64.efi'},
                {'source': 'boot.csv',
                 'target': 'EFI/ubuntu/bootx64.csv'},
                {'source': 'fb.efi',
                 'target': 'EFI/boot/fbx64.efi'},
                {'source': 'shim.efi.signed',
                 'target': 'EFI/boot/bootx64.efi'},
            ]
    elif partition.get('role') == 'system-boot':
        partition['update']['edition'] += args.edition_bump

yaml.safe_dump(content, stream=args.output)




import copy
import dash
from dash import Dash, html, Output, Input, dash_table, State, MATCH, ALL
import dash_bootstrap_components as dbc
import json
import os
import re
import sys

# To ensure this can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import query_features as qf

suppress_callback_exceptions = True
app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

retriever = None
timestamps = None
timestamp_options = []
coverage_matrix = {}
feature_options = [{"label": item, "value": item} for item in qf.KNOWN_FEATURES]
cached_duplicates = {}
cached_all_features_diff = {}
cached_feat_explore = {}
cached_feat_diff = {}


@app.callback(
    Output("error-with-retriever", "displayed"),
    Output({"type": "toggle-button", "index": ALL}, "disabled"),
    Output({"type": "timestamp-dropdown", "index": ALL}, "options"),
    Output("empty-div", "children"),
    Input("set-data-button", "n_clicks"),
    State("input-path", "value"),
)
def set_retriever(_, path):
    if not path:
        raise dash.exceptions.PreventUpdate

    if not os.path.exists(path):
        # First return true to display the error message.
        # Next, there are 5 toggle buttons and 6 timestamp dropdowns. 
        # Make all of the toggle buttons disabled and clear out
        # the options for all timestamp dropdowns.
        # Finally, return an empty output to signal the 
        # dcc.loading that this callback has finished.
        return True, [True] * 5, [[], [], [], [], [], []], []
    global retriever
    if os.path.isfile(path):
        with open(path, "r", encoding="utf-8") as f:
            retriever = qf.MongoRetriever(f)
    elif os.path.isdir(path):
        retriever = qf.DirRetriever(path)
    global timestamps
    timestamps = retriever.get_sorted_timestamps_and_systems()
    global timestamp_options
    timestamp_options = [
        {"label": item["timestamp"], "value": item["timestamp"]} for item in timestamps
    ]
    # First, return false so the error message is not displayed
    # Enable all 5 of the toggle buttons.
    # Populate all 6 timestamp dropdowns with the values.
    # Finally, return an empty output to signal the dcc.loading
    # that this callback has finished.
    return False, [False] * 5, [timestamp_options] * 6, []


def get_columns_from_list_of_dicts(features):
    i, _ = max(enumerate(features), key=lambda x: len(x[1]))
    return [{"name": key, "id": key} for key in features[i].keys()]


def make_dict_table_friendly(features):
    """
    For dictionary values that are themselves dictionaries,
    changes them to their string representation
    """
    processed = []
    for feature in features:
        feat_dict = {}
        for k, v in feature.items():
            feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
        processed.append(feat_dict)
    return processed


@app.callback(
    Output({"type": "collapse", "index": MATCH}, "is_open"),
    Input({"type": "toggle-button", "index": MATCH}, "n_clicks"),
    State({"type": "collapse", "index": MATCH}, "is_open"),
)
def toggle_collapse(n_clicks, is_open):
    if n_clicks:
        return not is_open
    return is_open


@app.callback(
    Output({"type": "systems-dropdown", "index": MATCH}, "options"),
    Input({"type": "timestamp-dropdown", "index": MATCH}, "value"),
)
def update_systems_dropdown(selected_timestamp):
    if not selected_timestamp:
        return []
    for item in timestamps:
        if item["timestamp"] == selected_timestamp:
            return [{"label": sys, "value": sys} for sys in item["systems"]]
    return []


@app.callback(
    Output("all-features-diff-table", "columns"),
    Output("all-features-diff-table", "data"),
    Input({"type": "timestamp-dropdown", "index": 1}, "value"),
)
def update_totals_table(selected_timestamp):
    if not selected_timestamp:
        return [], []

    diff_data = []
    for timestamp in timestamps:
        if timestamp["timestamp"] != selected_timestamp:
            continue
        for system in timestamp["systems"]:
            global cached_all_features_diff
            if (
                selected_timestamp in cached_all_features_diff
                and system in cached_all_features_diff[selected_timestamp]
            ):
                diff = cached_all_features_diff[selected_timestamp][system]
            else:
                diff = qf.diff_all_features(
                    retriever, selected_timestamp, system, False
                )
                if selected_timestamp not in cached_all_features_diff:
                    cached_all_features_diff[selected_timestamp] = {}
                cached_all_features_diff[selected_timestamp][system] = diff

            d = {key: len(value) for key, value in diff.items()}
            d["system"] = system
            diff_data.append(d)

    columns = [{"name": "System", "id": "system"}] + [
        {"name": key, "id": key} for key in qf.KNOWN_FEATURES
    ]

    return columns, diff_data


@app.callback(
    Output("all-features-diff-cell-data-container", "children"),
    Input("all-features-diff-table", "active_cell"),
    State("all-features-diff-table", "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 1}, "value"),
)
def display_column_details(active_cell, table_data, timestamp):
    if active_cell is None:
        return "Click a column cell to see details."

    row_idx = active_cell["row"]
    col_name = active_cell["column_id"]
    test = table_data[row_idx]

    features = cached_all_features_diff[timestamp][test["system"]][col_name]
    processed = []
    for feature in features:
        feat_dict = {}
        for k, v in feature.items():
            feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
        processed.append(feat_dict)

    table = dash_table.DataTable(
        data=processed,
        columns=get_columns_from_list_of_dicts(features),
        filter_action="native",
        sort_action="native",
        style_cell={
            "textAlign": "center",
            "maxWidth": "100%",
            "whiteSpace": "normal",
        },
        style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
    )

    return [html.H4(f"{test['system']} -- {col_name}"), table]


@app.callback(
    Output("coverage-diff-container", "children"),
    Input({"type": "timestamp-dropdown", "index": 2}, "value"),
    Input({"type": "systems-dropdown", "index": 2}, "value"),
    Input({"type": "systems-regex-input", "index": 2}, "value"),
    Input({"type": "timestamp-dropdown", "index": 3}, "value"),
    Input({"type": "systems-dropdown", "index": 3}, "value"),
    Input("remove-failed-switch", "on"),
    Input("only-same-switch", "on"),
    Input("match-snap-types", "on"),
)
def update_totals_table(ts_2, sys_2, sys_2_reg, ts_3, sys_3, remove_failed_value, only_same_value, match_snap_types):
    if not ts_2 or (not sys_2 and not sys_2_reg) or not ts_3 or not sys_3:
        return []
    
    global cached_feat_diff

    if not sys_2:
        for item in timestamps:
            if item["timestamp"] == ts_2:
                sys_list = [sys for sys in item["systems"]]
        try:
            pattern = re.compile(sys_2_reg)
            systems = [s for s in sys_list if pattern.search(s)]
            diff = {}
            cached_feat_diff = {}
            for system in systems:
                sys_diff = qf.diff(
                    retriever, ts_2, system, ts_3, sys_3, remove_failed_value, only_same_value, match_snap_types
                )
                diff = qf.union(diff, sys_diff)
                sys_test_diff = qf.diff_group_by_test(
                    retriever, ts_2, system, ts_3, sys_3, remove_failed_value, only_same_value, match_snap_types
                )
                for test_name, feat_dict in sys_test_diff.items():
                    if test_name not in cached_feat_diff:
                        cached_feat_diff[test_name] = feat_dict
                    else:
                        for feat_name, feat_list in feat_dict.items():
                            if feat_name not in cached_feat_diff[test_name]:
                                cached_feat_diff[test_name][feat_name] = feat_list
                            else:
                                for feat in feat_list:
                                    if feat not in cached_feat_diff[test_name][feat_name]:
                                        cached_feat_diff[test_name][feat_name].append(feat)
            i = 0
        except re.error:
            raise RuntimeError(f'regex {sys_2_reg} is invalid')

    if sys_2:
        diff = qf.diff(
            retriever, ts_2, sys_2, ts_3, sys_3, remove_failed_value, only_same_value, match_snap_types
        )
        cached_feat_diff = qf.diff_group_by_test(
            retriever, ts_2, sys_2, ts_3, sys_3, remove_failed_value, only_same_value, match_snap_types
        )

    tables = []
    column_data = [
        {"name": "test", "id": "test"},
        {"name":"# interfaces", "id":"# interfaces"},
        {"name":"# cmds", "id":"# cmds"},
        {"name":"# endpoints", "id":"# endpoints"},
        {"name":"# tasks", "id":"# tasks"},
        {"name":"# changes", "id":"# changes"},
        {"name":"# ensures", "id":"# ensures"},
        ]
    test_data = [
        {
            "test":test, 
            "# interfaces": len(features["interfaces"]) if "interfaces" in features else 0,
            "# cmds": len(features["cmds"]) if "cmds" in features else 0,
            "# endpoints": len(features["endpoints"]) if "endpoints" in features else 0,
            "# tasks": len(features["tasks"]) if "tasks" in features else 0,
            "# changes": len(features["changes"]) if "changes" in features else 0,
            "# ensures": len(features["ensures"]) if "ensures" in features else 0,
            } 
        for test, features in cached_feat_diff.items()
        ]
    table = dash_table.DataTable(
        id={"type": "coverage-diff-table", "index":0},
        data=test_data,
        columns=column_data,
        filter_action="native",
        sort_action="native",
        style_cell={
            "textAlign": "center",
            "maxWidth": "100%",
            "whiteSpace": "normal",
        },
        style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
    )
    tables.append(
        html.Div(
            [html.H4("Tests that contain the following missing features"), table],
            style={"maxWidth": "100%", "margin": "auto"},
        )
    )
    
    i = 1
    for feature_name, features in reversed(diff.items()):
        processed = []
        for feature in features:
            feat_dict = {}
            for k, v in feature.items():
                feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
            processed.append(feat_dict)
        table = dash_table.DataTable(
            id={"type": "coverage-diff-table", "index":i},
            data=processed,
            columns=get_columns_from_list_of_dicts(features),
            filter_action="native",
            sort_action="native",
            style_cell={
                "textAlign": "center",
                "maxWidth": "100%",
                "whiteSpace": "normal",
            },
            style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
        )
        tables.append(
            html.Div(
                [html.H4(feature_name), table],
                style={"maxWidth": "100%", "margin": "auto"},
            )
        )
        i=i+1

    return tables

@app.callback(
    Output("coverage-diff-modal-body", "children"),
    Output("coverage-diff-modal", "is_open"),
    Input({"type": "coverage-diff-table", "index": ALL}, "active_cell"),
    State({"type": "coverage-diff-table", "index": ALL}, "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 2}, "value"),
    State({"type": "systems-dropdown", "index": 2}, "value"),
    Input("match-snap-types", "on"),
)
def populate_tests_in_coverage_diff_cmds(active_cell, table_data, timestamp, system, match_snap_types):
    triggered = dash.callback_context.triggered_id
    if not active_cell or not table_data or not triggered or not 'index' in triggered or not active_cell[triggered["index"]] or not table_data[triggered["index"]]:
        raise dash.exceptions.PreventUpdate

    row_idx = active_cell[triggered["index"]]["row"]

    if triggered["index"] == 0:
        column_id = active_cell[triggered["index"]]["column_id"]
        if column_id == "test":
            return html.Div()
        
        test = table_data[triggered["index"]][row_idx]
        test_name = test["test"]
        feature_dict = cached_feat_diff[test_name]
        if column_id.split()[1] not in feature_dict:
            return html.Div()
        
        selected_feats = feature_dict[column_id.split()[1]]
        
        processed = []
        for selected_feat in selected_feats:
            feat_dict = {}
            for k, v in selected_feat.items():
                feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
            processed.append(selected_feat)
        table = dash_table.DataTable(
            data=processed,
            columns=get_columns_from_list_of_dicts(selected_feats),
            filter_action="native",
            sort_action="native",
            style_cell={
                "textAlign": "center",
                "maxWidth": "100%",
                "whiteSpace": "normal",
            },
            style_table={"overflowX": "auto", "maxWidth": "100%", "minWidth": "600px", "margin": "auto"},
        )
        return html.Div(
            [html.H4(f"{test_name} --- {column_id.split()[1]}", style={"textAlign": "center"}), table],
            style={"maxWidth": "100%", "margin": "auto"},
        ), True


    feature = copy.deepcopy(table_data[triggered["index"]][row_idx])
    for k, v in feature.items():
        try:
            # Features with a snap type list like tasks and changes have been stringified
            # for visualization purposes in the GUI. Change them back to lists of strings.
            # All other field values that are not valid json can remain unaltered.
            feature[k] = json.loads(v)
        except:
            pass

    results = qf.find_feat(retriever, timestamp, feature, remove_failed=False, system=system, match_snap_types=match_snap_types)
    table = dash_table.DataTable(
        data=get_task_list_from_dict(results),
        columns=[{"name":"suite","id":"suite"}, {"name":"test","id":"test"}, {"name":"variant","id":"variant"}],
        filter_action="native",
        sort_action="native",
        style_cell={
            "textAlign": "center",
            "maxWidth": "100%",
            "whiteSpace": "normal",
        },
        style_table={"overflowX": "auto", "maxWidth": "100%", "minWidth": "600px", "margin": "auto"},
    )
    return html.Div(
        [html.H4(f"{system} --- {feature}", style={"textAlign": "center"}), table],
        style={"maxWidth": "100%", "margin": "auto"},
    ), True


@app.callback(
    Output({"type": "timestamp-dropdown", "index": 2}, "value"),
    Output({"type": "systems-dropdown", "index": 2}, "value"),
    Output({"type": "timestamp-dropdown", "index": 3}, "value"),
    Output({"type": "systems-dropdown", "index": 3}, "value"),
    Input("switch-button", "n_clicks"),
    State({"type": "timestamp-dropdown", "index": 2}, "value"),
    State({"type": "systems-dropdown", "index": 2}, "value"),
    State({"type": "timestamp-dropdown", "index": 3}, "value"),
    State({"type": "systems-dropdown", "index": 3}, "value"),
)
def switch_dropdown_values(n_clicks, ts2, sys2, ts3, sys3):
    if n_clicks is None or n_clicks == 0:
        raise dash.exceptions.PreventUpdate

    return ts3, sys3, ts2, sys2


@app.callback(
    Output("suite-dropdown-filter", "options"),
    Output("task-dropdown-filter", "options"),
    Output("variant-dropdown-filter", "options"),
    Input({"type": "timestamp-dropdown", "index": 4}, "value"),
)
def update_systems_dropdown(selected_timestamp):
    if not selected_timestamp:
        raise dash.exceptions.PreventUpdate
    task_list = qf.task_list(retriever, selected_timestamp)
    suites = set([task.suite for task in task_list])
    tasks = set([task.task_name for task in task_list])
    variants = set([task.variant for task in task_list])
    return list(suites), list(tasks), list(variants)


@app.callback(
    Output("coverage-matrix-table", "columns"),
    Output("coverage-matrix-table", "data"),
    Input({"type": "timestamp-dropdown", "index": 4}, "value"),
    Input("coverage-remove-failed-switch", "on"),
    Input("suite-dropdown-filter", "value"),
    Input("task-dropdown-filter", "value"),
    Input("variant-dropdown-filter", "value"),
)
def create_coverage_matrix(timestamp, remove_failed, suite, task, variant):
    if not dash.callback_context.triggered:
        raise dash.exceptions.PreventUpdate

    systems = None
    for ts in timestamps:
        if ts["timestamp"] == timestamp:
            systems = ts["systems"]
    if not systems:
        return [], []

    columns = [{"name": "System", "id": "system"}] + [
        {"name": key, "id": key} for key in qf.KNOWN_FEATURES
    ]
    global coverage_matrix
    coverage_matrix[timestamp] = [
        {"system": system, **{key: 0 for key in qf.KNOWN_FEATURES}}
        for system in systems
    ]
    matrix = [
        {"system": system, **{key: 0 for key in qf.KNOWN_FEATURES}}
        for system in systems
    ]
    for i, system in enumerate(systems):
        feats = qf.feat_sys(
            retriever, timestamp, system, remove_failed, suite, task, variant
        )
        coverage_matrix[timestamp][i].update(feats)
        for feature in qf.KNOWN_FEATURES:
            matrix[i][feature] = len(feats[feature])

    return columns, matrix


@app.callback(
    Output("cell-data-container", "children"),
    Input("coverage-matrix-table", "active_cell"),
    State("coverage-matrix-table", "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 4}, "value"),
)
def display_cell_data(active_cell, table_data, timestamp):
    if not active_cell or not table_data or not timestamp:
        return "Click on a cell to see feature data"

    row_idx = active_cell["row"]
    col_idx = active_cell["column_id"]

    # Get the system name from the row data
    system = table_data[row_idx]["system"]

    system_data = next(
        (item for item in coverage_matrix[timestamp] if item["system"] == system), None
    )
    if not system_data:
        return "No data found for the selected cell."

    features = system_data[col_idx]
    if len(features) == 0:
        return "No data found for the selected cell."

    table = dash_table.DataTable(
        id="coverage-feature-table",
        data=make_dict_table_friendly(features),
        columns=get_columns_from_list_of_dicts(features),
        filter_action="native",
        sort_action="native",
        style_cell={
            "textAlign": "center",
            "maxWidth": "100%",
            "whiteSpace": "normal",
        },
        style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
    )
    return html.Div(
        [html.H4(f"{system} ---- {col_idx}:", style={"textAlign": "center"}), table],
        style={"maxWidth": "100%", "margin": "auto"},
    )



@app.callback(
    Output("coverage-modal-body", "children"),
    Output("coverage-modal", "is_open"),
    Input("coverage-feature-table", "active_cell"),
    State("coverage-feature-table", "derived_viewport_data"),
    State("coverage-matrix-table", "active_cell"),
    State("coverage-matrix-table", "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 4}, "value"),
    State("match-snap-types", "on"),
)
def populate_tests_in_coverage_diff_cmds(active_cell, table_data, system_active_cell, system_table_data, timestamp, match_snap_types):
    if not active_cell:
        raise dash.exceptions.PreventUpdate

    # Get the system name from the system matrix data
    system = system_table_data[system_active_cell["row"]]["system"]

    row_idx = active_cell["row"]
    feature = copy.deepcopy(table_data[row_idx])
    for k, v in feature.items():
        try:
            # Features with a snap type list like tasks and changes have been stringified
            # for visualization purposes in the GUI. Change them back to lists of strings.
            # All other field values that are not valid json can remain unaltered.
            feature[k] = json.loads(v)
        except:
            pass

    results = qf.find_feat(retriever, timestamp, feature, remove_failed=False, system=system, match_snap_types=match_snap_types)
    table = dash_table.DataTable(
        data=get_task_list_from_dict(results),
        columns=[{"name":"suite","id":"suite"}, {"name":"test","id":"test"}, {"name":"variant","id":"variant"}],
        filter_action="native",
        sort_action="native",
        style_cell={
            "textAlign": "center",
            "maxWidth": "100%",
            "whiteSpace": "normal",
        },
        style_table={"overflowX": "auto", "maxWidth": "100%", "minWidth": "600px", "margin": "auto"},
    )
    return html.Div(
        [html.H4(f"{system} --- {feature}", style={"textAlign": "center"}), table],
        style={"maxWidth": "100%", "margin": "auto"},
    ), True


@app.callback(
    Output("duplicate-table", "columns"),
    Output("duplicate-table", "data"),
    Input({"type": "timestamp-dropdown", "index": 5}, "value"),
    Input({"type": "systems-dropdown", "index": 5}, "value"),
)
def calculate_duplicate_systems(timestamp, system):

    if not timestamp or not system:
        return [], []

    global cached_duplicates
    if timestamp in cached_duplicates and system in cached_duplicates[timestamp]:
        duplicates = cached_duplicates[timestamp][system]
    else:
        duplicates = qf.dup(retriever, timestamp, system, False)
        if timestamp in cached_duplicates:
            cached_duplicates[timestamp][system] = duplicates
        else:
            cached_duplicates[timestamp] = {system: duplicates}

    columns = [
        {"name": "suite", "id": "suite"},
        {"name": "task", "id": "task"},
        {"name": "variant", "id": "variant"},
    ]

    rows = [
        {"suite": d.suite, "task": d.task_name, "variant": d.variant}
        for d in duplicates
    ]

    return columns, rows


@app.callback(
    Output("cell-duplicate-container", "children"),
    Input("duplicate-table", "active_cell"),
    State("duplicate-table", "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 5}, "value"),
    State({"type": "systems-dropdown", "index": 5}, "value"),
)
def display_duplicate_cell_data(active_cell, table_data, timestamp, system):
    if not active_cell or not table_data or not timestamp:
        return "Click on a cell to see feature data"

    row_idx = active_cell["row"]

    test = table_data[row_idx]

    features = qf.feat_sys(
        retriever,
        timestamp,
        system,
        False,
        suite=test["suite"],
        task=test["task"],
        variant=test["variant"],
    )

    tables = [
        html.H4(
            f"{system}:{str(qf.TaskIdVariant(test['suite'], test['task'], test['variant']))}",
            style={"textAlign": "center"},
        )
    ]
    for feature_name, feature_data in features.items():
        processed = []
        for feature in feature_data:
            feat_dict = {}
            for k, v in feature.items():
                feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
            processed.append(feat_dict)
        table = dash_table.DataTable(
            data=processed,
            columns=get_columns_from_list_of_dicts(feature_data),
            filter_action="native",
            sort_action="native",
            style_cell={
                "textAlign": "center",
                "maxWidth": "100%",
                "whiteSpace": "normal",
            },
            style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
        )
        tables.append(
            html.Div(
                [html.H4(feature_name), table],
                style={"maxWidth": "100%", "margin": "auto"},
            )
        )

    return tables


def get_task_list_from_dict(tests: dict[str, qf.TaskIdVariant]) -> list[dict]:
    processed = []
    for _, test_list in tests.items():
        for test in test_list:
            d = {"suite": test.suite, "test": test.task_name, "variant": test.variant}
            if d not in processed:
                processed.append(d)
    return processed


@app.callback(
    Output("explore-by-feature-table", "columns"),
    Output("explore-by-feature-table", "data"),
    Input({"type": "timestamp-dropdown", "index": 6}, "value"),
    Input("features-dropdown", "value"),
    Input("add-frequency-numbers-switch", "on"),
)
def populate_feature_table(timestamp, selected_feature, add_freq):
    if not timestamp or not selected_feature:
        return [], []

    global cached_feat_explore
    if (
        add_freq
        and timestamp in cached_feat_explore
        and selected_feature in cached_feat_explore[timestamp]
    ):
        return (
            cached_feat_explore[timestamp][selected_feature]["cols"],
            cached_feat_explore[timestamp][selected_feature]["processed"],
        )

    features = retriever.get_all_features(timestamp)
    feature_data = features[selected_feature]
    processed = []
    for feature in feature_data:
        feat_dict = {}
        if add_freq:
            feat_dict["number-of-tests-with-feature"] = len(
                get_task_list_from_dict(
                    qf.find_feat(retriever, timestamp, feature, False)
                )
            )
        for k, v in feature.items():
            feat_dict[k] = json.dumps(v) if isinstance(v, list) else v
        processed.append(feat_dict)
    cols = get_columns_from_list_of_dicts(feature_data)
    if add_freq:
        cols = [{"name": "Number of tests with feature", "id": "number-of-tests-with-feature"}] + cols

    if add_freq:
        if timestamp not in cached_feat_explore:
            cached_feat_explore[timestamp] = {}
        cached_feat_explore[timestamp][selected_feature] = {
            "cols": cols,
            "processed": processed,
        }
    return cols, processed


@app.callback(
    Output("explore-by-feature-data-container", "children"),
    Input("explore-by-feature-table", "active_cell"),
    State("explore-by-feature-table", "derived_viewport_data"),
    State({"type": "timestamp-dropdown", "index": 6}, "value"),
    State("features-dropdown", "value"),
)
def update_test_list(active_cell, table_data, timestamp, selected_feature):
    if not active_cell or not table_data or not timestamp or not selected_feature:
        return "Click on a cell to see tests"

    row_idx = active_cell["row"]

    feature = copy.deepcopy(table_data[row_idx])
    if "number-of-tests-with-feature" in feature:
        del feature["number-of-tests-with-feature"]

    tests = qf.find_feat(retriever, timestamp, feature, False)

    processed = []
    sys_dict = {}
    for sys, test_list in tests.items():
        for test in test_list:
            d = {"suite": test.suite, "test": test.task_name, "variant": test.variant}
            s = json.dumps(d)
            if s not in sys_dict:
                sys_dict[s] = []
            sys_dict[s].append(sys)
            if d not in processed:
                processed.append(d)

    for p in processed:
        s = json.dumps(p)
        p["systems"] = "\n".join(sorted(sys_dict[s]))
        p["systems"] = p["systems"].replace('-', '_')

    return [
        html.H4(f"Tests that contain feature {feature}"),
        dash_table.DataTable(
            data=processed,
            columns=[
                {"name": "suite", "id": "suite"},
                {"name": "test", "id": "test"},
                {"name": "variant", "id": "variant"},
                {"name": "systems", "id": "systems"},
            ],
            filter_action="native",
            sort_action="native",
            style_cell={
                "textAlign": "center",
                "maxWidth": "auto",
                "whiteSpace": "normal",
            },
            style_data_conditional=[
                {
                    "if": {"column_id": "systems"},
                    "whiteSpace": "pre-line",
                    "maxWidth": "auto",
                    "width": "auto",
                }
            ],
            style_table={"overflowX": "auto", "maxWidth": "100%", "margin": "auto"},
        ),
    ]


#!/usr/bin/env python3

import sys
import argparse

import yaml


def parse_arguments():
    parser = argparse.ArgumentParser(description="pc gadget yaml generator for test")
    parser.add_argument(
        "gadgetyaml", type=argparse.FileType("r"), help="path to gadget.yaml input file"
    )
    return parser.parse_args()


def must_find_struct(structs, name):
    found = [s for s in structs if s["name"] == name]
    if len(found) != 1:
        raise RuntimeError("no structure with name {} among: {}".format(name, structs))
    return found[0]


def bump_update_edition(update):
    if update is None:
        return {"edition": 1}
    if "edition" not in update:
        update["edition"] = 1
    else:
        update["edition"] += 1
    return update


def main(opts):
    gadget_yaml = yaml.safe_load(opts.gadgetyaml)

    structs = gadget_yaml["volumes"]["pc"]["structure"]
    ubuntu_seed = must_find_struct(structs, "ubuntu-seed")
    ubuntu_seed["update"] = bump_update_edition(ubuntu_seed.get("update"))

    # trigger an update, so that device lookup is performed
    mbr = must_find_struct(structs, "mbr")
    mbr["update"] = bump_update_edition(mbr.get("update"))

    yaml.dump(gadget_yaml, stream=sys.stdout, indent=2, default_flow_style=False)


if __name__ == "__main__":
    opts = parse_arguments()
    main(opts)




import argparse
from io import StringIO
import json
import os
import tempfile
from typing import Any
import unittest
from unittest.mock import Mock, patch
import sys

# To ensure the unit test can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import featcomposer
from features import *


class TestCompose(unittest.TestCase):
    @staticmethod
    def get_features(msg: str) -> dict[str, Any]:
        features = {}
        features['cmds'] = [Cmd(cmd=msg), Cmd(cmd=f'msg msg')]
        features['endpoints'] = [Endpoint(method='POST', path=msg)]
        features['interfaces'] = [Interface(name=msg)]
        features['tasks'] = [
            Task(kind=msg, snap_type='snap', last_status=Status.done)]
        features['changes'] = [Change(kind=msg, snap_type='snap')]
        return features

    @staticmethod
    def get_json(suite: str, task: str, variant: str, success: str, msg: str) -> TaskFeatures:
        features = TestCompose.get_features(msg)
        return TaskFeatures(
            suite=suite,
            task_name=task,
            variant=variant,
            success=success,
            cmds=features['cmds'],
            endpoints=features['endpoints'],
            interfaces=features['interfaces'],
            tasks=features['tasks'],
            changes=features['changes']
        )

    @staticmethod
    def write_task(filepath: str, msg: str) -> None:
        with open(filepath, 'w') as f:
            json.dump(TestCompose.get_features(msg), f)

    def test_compose(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            task1variant1 = os.path.join(
                tmpdir, 'backend:system.version:path--to--task1:variant1.json')
            task2 = os.path.join(tmpdir, 'backend:system.version:path--to--task2')
            TestCompose.write_task(task1variant1, 'task1variant1')
            TestCompose.write_task(task2, 'task2')
            systems = featcomposer.get_system_list(tmpdir)
            self.assertEqual(1, len(systems))
            composed = featcomposer.compose_system(tmpdir, systems.pop(),
                                                       'backend:system.version:path/to/task1:variant1 backend:system.version:another/task2',
                                                       ['e = 1 ', 'f = 2 '], ['1 ', ' 2', ' 3'])
            assert len(composed) == 5
            assert 'schema_version' in composed and composed['schema_version'] == '0.0.0'
            assert 'system' in composed and composed['system'] == 'backend:system.version'
            assert 'scenarios' in composed and len(composed['scenarios']) == 3
            assert '1' in composed['scenarios']
            assert '2' in composed['scenarios']
            assert '3' in composed['scenarios']
            assert 'env_variables' in composed and len(composed['env_variables']) == 2
            assert {'name': 'e', 'value': '1'} in composed['env_variables']
            assert {'name': 'f', 'value': '2'} in composed['env_variables']
            assert 'tests' in composed and len(composed['tests']) == 2
            assert TestCompose.get_json('path/to', 'task1', 'variant1', False, 'task1variant1') in composed['tests']
            assert TestCompose.get_json('path/to', 'task2', '', True, 'task2') in composed['tests']

    @patch('argparse.ArgumentParser.parse_args')
    def test_compose_features(self, parse_args_mock: Mock):
        with tempfile.TemporaryDirectory() as tmpdir:
            res = os.path.join(tmpdir, 'res')
            os.mkdir(res)
            sys1test1 = TestCompose.get_features('sys1test1')
            sys1test2 = TestCompose.get_features('sys1test2')
            sys2test1 = TestCompose.get_features('sys2test1')
            sys2test3 = TestCompose.get_features('sys2test3')

            with open(os.path.join(res, 'backend:system1:tests--test1'), mode='w', encoding='utf-8') as f:
                json.dump(sys1test1, f)
            with open(os.path.join(res, 'backend:system1:tests--test2:variant1'), mode='w', encoding='utf-8') as f:
                json.dump(sys1test2, f)
            with open(os.path.join(res, 'backend:system2:tests--test1'), mode='w', encoding='utf-8') as f:
                json.dump(sys2test1, f)
            with open(os.path.join(res, 'backend:system2:tests--test3'), mode='w', encoding='utf-8') as f:
                json.dump(sys2test3, f)
            out = os.path.join(tmpdir, 'out')
            os.mkdir(out)
            failed=StringIO('backend:system1:tests/test1 backend:system1:tests/test2:variant1 backend:system2:tests/test3')
            parse_args_mock.return_value = argparse.Namespace(
                    dir=res,
                    output=out,
                    failed_tests=failed,
                    run_attempt=1,
                    replace_old_runs=False,
                    env_variables='',
                    scenarios=''
                )
            featcomposer.main()

            def check_test_equal(expected, actual, succeeded):
                assert succeeded == actual['success']
                for k in expected.keys():
                    assert k in actual
                    assert actual[k] == expected[k]

            with open(os.path.join(out, 'backend:system1_1.json'), mode='r', encoding='utf-8') as f:
                sys1 = json.load(f)
                for test in sys1['tests']:
                    assert 'test1' in test['task_name'] or 'test2' in test['task_name']
                    if test['task_name'] == 'test1':
                        check_test_equal(sys1test1, test, False)
                    if test['task_name'] == 'test2':
                        check_test_equal(sys1test2, test, False)
            with open(os.path.join(out, 'backend:system2_1.json'), mode='r', encoding='utf-8') as f:
                sys1 = json.load(f)
                for test in sys1['tests']:
                    assert 'test1' in test['task_name'] or 'test3' in test['task_name']
                    if test['task_name'] == 'test1':
                        check_test_equal(sys2test1, test, True)
                    if test['task_name'] == 'test3':
                        check_test_equal(sys2test3, test, False)


class TestReplace(unittest.TestCase):

    def test_replace(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            original = os.path.join(tmpdir, 'my:system.version_1')
            rerun = os.path.join(tmpdir, 'my:system.version_2.json')
            run_once = os.path.join(tmpdir, 'my:other-system.version_1.json')
            original_json = {'system.version': 'my:system.version', 'tests': [{'task_name': 'task1', 'suite': 'my/suite', 'variant': '', 'success': False, 'cmds': [{'cmd': 'original run'}]},
                                                              {'task_name': 'task2', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'original run'}]}]}
            rerun_json = {'system.version': 'my:system.version', 'tests': [
                {'task_name': 'task1', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'rerun 1'}, {'cmd': 'another'}]},
                {'task_name': 'task3', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'rerun 2'}]}]}
            run_once_json = {'system.version': 'my:other-system.version', 'tests': [
                {'task_name': 'task', 'suite': 'my/suite', 'variant': 'v1', 'success': True}]}
            with open(original, 'w') as f:
                json.dump(original_json, f)
            with open(rerun, 'w') as f:
                json.dump(rerun_json, f)
            with open(run_once, 'w') as f:
                json.dump(run_once_json, f)
            output_dir = 'replaced'
            os.makedirs(os.path.join(tmpdir, output_dir), exist_ok=True)
            featcomposer.replace_old_runs(
                tmpdir, os.path.join(tmpdir, output_dir))
            self.assertEqual(
                2, len(os.listdir(os.path.join(tmpdir, output_dir))))
            with open(os.path.join(tmpdir, output_dir, 'my:system.version.json'), 'r') as f:
                actual = json.load(f)
                assert len(actual) == 2
                assert 'system.version' in actual and actual['system.version'] == 'my:system.version'
                assert 'tests' in actual and len(actual['tests']) == 3
                assert {'task_name': 'task1', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'rerun 1'}, {'cmd': 'another'}]} in actual['tests']
                assert {'task_name': 'task2', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'original run'}]} in actual['tests']
                assert {'task_name': 'task3', 'suite': 'my/suite', 'variant': '', 'success': True, 'cmds': [{'cmd': 'rerun 2'}]} in actual['tests']
            with open(os.path.join(tmpdir, output_dir, 'my:other-system.version.json'), 'r') as f:
                actual = json.load(f)
                assert len(actual) == 2
                assert 'system.version' in actual and actual['system.version'] == run_once_json['system.version']
                assert 'tests' in actual and actual['tests'] == run_once_json['tests']

    def test_replace_missing_original(self):
        filenames = ["f_1", "f_2", "f_3", "g_2", "g_3", "h_5", "i_1"]
        originals, reruns = featcomposer._get_original_and_rerun_list(filenames)
        assert "f_1" in originals
        assert "g_2" in originals

        assert "f_2" in reruns
        assert "f_3" in reruns
        assert "g_3" in reruns


if __name__ == '__main__':
    unittest.main()



import unittest

from common import (
    LockOpTrace,
)


class TestLockOpTrace(unittest.TestCase):
    def setUp(self):
        self.lines = ["line 1\n", "line 2\n", "line 3\n"]
        self.trace = LockOpTrace(self.lines)

    def test_get_trace_lines(self):
        self.assertEqual(self.trace.get_trace_lines(), self.lines)

    def test_match(self):
        self.assertTrue(self.trace.match("line 2"))
        self.assertFalse(self.trace.match("line 4"))

    def test_str(self):
        self.assertEqual(str(self.trace), "line 1\nline 2\nline 3")

    def test_eq(self):
        other_trace = LockOpTrace(self.lines)
        self.assertEqual(self.trace, other_trace)



#!/usr/bin/env python3

from __future__ import annotations


class LockOpTrace:
    """
    LockOpTrace: Represents a Lock Operation trace in the log file.
    It holds the lines for the trace and allows matching against a part of
    the trace.
    """

    def __init__(self, lines: list[str]):
        self.lines = lines
        self.hash = hash(str(self))

    def get_trace_lines(self) -> list[str]:
        return self.lines

    def match(self, part: str) -> bool:
        for line in self.lines:
            if part in line:
                return True

        return False

    def print(self):
        print("-" * 20 + "TRACE" + "-" * 20)
        print("")
        print(self)
        print("")

    def __str__(self) -> str:
        return "".join(self.lines).rstrip()

    def __hash__(self) -> int:
        return self.hash

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LockOpTrace):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.hash == other.hash


"""
Generic function used to iterate a list of strings until it is found an
element which starts with the prefix passed as parameter. The function
returns the index for the first match, and -1 in case there isn't any.
"""


def get_next_match(lines: list[str], start: int, prefix: str) -> int:
    for current_line in range(start + 1, len(lines)):
        if lines[current_line].startswith(prefix):
            return current_line
    return -1



#!/usr/bin/env python3

# patch-dl calls patchelf on ELF binaries with an interpreter to update
# the path of that interpreter.

import argparse
import os
import shutil
import subprocess
import tempfile
import logging

from elftools.elf.elffile import ELFFile


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="patch ELF binaries to request a specific interpreter",
    )
    craft_prime = os.environ.get("CRAFT_PRIME")
    parser.add_argument(
        "--prime",
        nargs=1,
        default=craft_prime,
        required=(craft_prime is None),
        help="snapcraft part priming location",
    )

    craft_part_install = os.environ.get("CRAFT_PART_INSTALL")
    parser.add_argument(
        "--install",
        nargs=1,
        default=craft_part_install,
        required=(craft_part_install is None),
        help="snapcraft part install location",
    )

    parser.add_argument("interp", help="ELF interpreter to use")

    return parser.parse_args()


def is_shared_exec(path: str) -> bool:
    with open(path, "rb") as f:
        if f.read(4) != b"\x7fELF":
            return False
        f.seek(0, 0)
        elf = ELFFile(f)
        for segment in elf.iter_segments():
            # TODO: use iter_segments(type='PT_INTERP')
            if segment["p_type"] == "PT_INTERP":
                return True
        return False


def main(args) -> None:
    owned_executables = []

    for dirpath, _, filenames in os.walk(args.install):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            if os.path.islink(path):
                continue
            if not is_shared_exec(path):
                continue
            logging.debug("found owned ELF binary: %s", path)
            rel = os.path.relpath(path, args.install)
            # Now we need to know if the file in $CRAFT_PRIME is actually
            # owned by the current part and see if it is hard-linked to a
            # corresponding file in $CRAFT_PART_INSTALL.
            #
            # Even if we break the hard-links before, subsequent builds will
            # re-introduce the hard-links in `crafctl default` call in the
            # `override-prime`.
            prime_path = os.path.join(args.prime, rel)
            install_st = os.lstat(path)
            prime_st = os.lstat(path)
            if install_st.st_dev != prime_st.st_dev:
                continue
            if install_st.st_ino != prime_st.st_ino:
                continue
            owned_executables.append(prime_path)

    for path in owned_executables:
        # Because files in $CRAFT_PRIME, $CRAFT_STAGE, and $CRAFT_PART_INSTALL are hard-linked,
        # we need to copy the file first to avoid writing back to the content of other directories.
        with tempfile.NamedTemporaryFile(
            dir=os.path.dirname(path), prefix=f"{os.path.basename(path)}-"
        ) as f:
            with open(path, "rb") as orig:
                shutil.copyfileobj(orig, f)
            f.flush()
            logging.info("patching ELF binary %s", path)
            subprocess.run(
                ["patchelf", "--set-interpreter", args.interp, f.name], check=True
            )
            shutil.copystat(path, f.name)
            os.unlink(path)
            os.link(f.name, path)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main(parse_arguments())



#!/usr/bin/env python3
import sys
from gi.repository import GLib
import dbus
import dbus.service

from dbus.mainloop.glib import DBusGMainLoop

DBusGMainLoop(set_as_default=True)


class DBusProvider(dbus.service.Object):
    def __init__(self, bus):
        bus_name = dbus.service.BusName("com.dbustest.HelloWorld", bus=bus)
        dbus.service.Object.__init__(self, bus_name, "/com/dbustest/HelloWorld")

    @dbus.service.method(dbus_interface="com.dbustest.HelloWorld", out_signature="s")
    def SayHello(self):
        return "hello world"


if __name__ == "__main__":
    if sys.argv[1] == "system":
        bus = dbus.SystemBus()
    elif sys.argv[1] == "session":
        bus = dbus.SessionBus()
    else:
        print("unknown bus: %s", sys.argv[1:])
        sys.exit(1)
    DBusProvider(bus)
    loop = GLib.MainLoop()
    loop.run()



#!/usr/bin/env python3

import sys
import argparse

import yaml


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="pc gadget yaml variant generator for test"
    )
    parser.add_argument(
        "--system-seed", action="store_true", help="also modify system-seed"
    )
    parser.add_argument(
        "--system-bios", action="store_true", help="also modify system-bios"
    )
    parser.add_argument(
        "gadgetyaml", type=argparse.FileType("r"), help="path to gadget.yaml input file"
    )
    parser.add_argument("variant", help="test data variant", choices=["v1", "v2"])
    return parser.parse_args()


def match_name(name):
    return lambda struct: struct.get("name", "") == name


def match_role_with_fallback(role):
    def match(struct):
        if "role" in struct:
            struct_role = struct["role"]
        elif "filesystem-label" in struct:
            # fallback to filesystem-label
            struct_role = struct["filesystem-label"]
        else:
            return False
        return role == struct_role

    return match


def must_find_struct(structs, matcher):
    found = [s for s in structs if matcher(s)]
    if len(found) != 1:
        raise RuntimeError("found {} matches among: {}".format(len(found), structs))
    return found[0]


def may_find_struct(structs, matcher):
    found = [s for s in structs if matcher(s)]
    if len(found) != 1:
        return None
    return found[0]


def bump_update_edition(update):
    if update is None:
        return {"edition": 1}
    if "edition" not in update:
        update["edition"] = 1
    else:
        update["edition"] += 1
    return update


def make_v1(doc, system_seed):
    # add new files to 'EFI System' partition, add new image file to 'BIOS
    # Boot', bump update edition for both
    structs = doc["volumes"]["pc"]["structure"]
    # "EFI System" in UC16/UC18, or just system-boot in UC20
    efisystem = must_find_struct(structs, match_role_with_fallback("system-boot"))
    biosboot = may_find_struct(structs, match_name("BIOS Boot"))

    # from UC16/UC18 gadgets:
    #
    # - name: EFI System
    #   (not)type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
    #   filesystem: vfat
    #   filesystem-label: system-boot
    #   size: 50M
    #   content:
    #     - source: grubx64.efi
    #       target: EFI/boot/grubx64.efi
    #     - source: shim.efi.signed
    #       target: EFI/boot/bootx64.efi
    #     - source: mmx64.efi
    #       target: EFI/boot/mmx64.efi
    #     - source: grub.cfg
    #       target: EFI/ubuntu/grub.cfg
    #     # drop a new file
    #     - source: foo.cfg
    #       target: foo.cfg
    #  update:
    #      edition: 1
    efisystem["content"].append({"source": "foo.cfg", "target": "foo.cfg"})
    efisystem["update"] = bump_update_edition(efisystem.get("update"))

    # - name: BIOS Boot
    #   (not)type: DA,21686148-6449-6E6F-744E-656564454649
    #   size: 1M
    #   offset: 1M
    #   offset-write: mbr+92
    #   content:
    #     - image: pc-core.img
    #     # write new content right after the previous one
    #     - image: foo.img
    #   update:
    #       edition: 1
    if biosboot and "content" in biosboot:
        biosboot["content"].append({"image": "foo.img"})
        biosboot["update"] = bump_update_edition(biosboot.get("update"))

    if system_seed:
        # from UC20 gadget:
        #
        # - name: ubuntu-seed
        #   role: system-seed
        #   filesystem: vfat
        #   # UEFI will boot the ESP partition by default first
        #   (not)type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
        #   size: 1200M
        #   update:
        #     edition: 2
        #   content:
        #     - source: grubx64.efi
        #       target: EFI/boot/grubx64.efi
        #     - source: shim.efi.signed
        #       target: EFI/boot/bootx64.efi
        #     - source: grub-recovery.conf
        #       target: EFI/ubuntu/grub.cfg
        systemseed = must_find_struct(structs, match_role_with_fallback("system-seed"))
        systemseed["content"].append(
            {"source": "foo-seed.cfg", "target": "foo-seed.cfg"}
        )
        systemseed["update"] = bump_update_edition(systemseed.get("update"))

    return doc


def make_v2(doc, system_seed, system_bios):
    # appply v1, add more new files to 'EFI System' partition, preserve one of
    # the updated files, to 'BIOS Boot', bump update edition for both

    doc = make_v1(doc, system_seed)

    structs = doc["volumes"]["pc"]["structure"]
    efisystem = must_find_struct(structs, match_role_with_fallback("system-boot"))
    if system_bios:
        biosboot = must_find_struct(structs, match_name("BIOS Boot"))
    else:
        biosboot = may_find_struct(structs, match_name("BIOS Boot"))

    # - name: EFI System
    #   (not)type: EF,C12A7328-F81F-11D2-BA4B-00A0C93EC93B
    #   filesystem: vfat
    #   filesystem-label: system-boot
    #   size: 50M
    #   content:
    #     - source: grubx64.efi
    #       target: EFI/boot/grubx64.efi
    #     - source: shim.efi.signed
    #       target: EFI/boot/bootx64.efi
    #     - source: mmx64.efi
    #       target: EFI/boot/mmx64.efi
    #     - source: grub.cfg
    #       target: EFI/ubuntu/grub.cfg
    #     # drop a new file
    #     - source: foo.cfg
    #       target: foo.cfg
    #     - source: bar.cfg
    #       target: bar.cfg
    #  update:
    #      edition: 2
    #      preserve: [foo.cfg, bar.cfg]
    efisystem["content"].append({"source": "bar.cfg", "target": "bar.cfg"})
    efisystem["update"] = bump_update_edition(efisystem.get("update"))
    efisystem["update"]["preserve"] = efisystem["update"].get("preserve", []) + [
        "foo.cfg",
        "bar.cfg",
    ]
    # - name: BIOS Boot
    #   (not)type: DA,21686148-6449-6E6F-744E-656564454649
    #   size: 1M
    #   offset: 1M
    #   offset-write: mbr+92
    #   content:
    #     - image: pc-core.img
    #     # content is modified again
    #     - image: foo.img
    #   update:
    #       edition: 2
    if system_bios: 
        biosboot["update"] = bump_update_edition(biosboot.get("update"))

    if system_seed:
        # only in UC20 gadgets
        systemseed = must_find_struct(structs, match_role_with_fallback("system-seed"))
        # we already appended foo-boot.cfg, bump the edition so that it gets
        # updated
        systemseed["update"] = bump_update_edition(systemseed.get("update"))

    return doc


def main(opts):
    doc = yaml.safe_load(opts.gadgetyaml)

    if opts.variant == "v1":
        make_v1(doc, opts.system_seed)
    elif opts.variant == "v2":
        make_v2(doc, opts.system_seed, opts.system_bios)

    yaml.dump(doc, sys.stdout)


if __name__ == "__main__":
    opts = parse_arguments()
    main(opts)



#!/usr/bin/python3

from io import StringIO
import os
import unittest

fake_news_md = """
# New in snapd 2.60.3:
* Fix bug in the "private" plug attribute of the shared-memory
  interface that can result in a crash when upgrading from an
  old version of snapd.
* Fix missing integration of the /etc/apparmor.d/tunables/home.d/
  apparmor to support non-standard home directories
"""

expected_changelog_entry = """
    - Fix bug in the "private" plug attribute of the shared-memory
      interface that can result in a crash when upgrading from an old
      version of snapd.
    - Fix missing integration of the /etc/apparmor.d/tunables/home.d/
      apparmor to support non-standard home directories
"""[
    1:
]


class TestChangelogReadNewsMd(unittest.TestCase):
    def setUp(self):
        try:
            # check whether python3-debian is installed, the tests could be run
            # on systems where the module is not package or otherwise available
            import changelog  # noqa: F401
        except ModuleNotFoundError as err:
            if err.name == "debian":
                raise unittest.SkipTest("skip test due to missing 'debian' module")
            raise

        self.news_md = StringIO(fake_news_md)

    def test_happy(self):
        import changelog

        changelog_entry = changelog.read_changelogs_news_md(self.news_md, "2.60.3")
        self.assertEqual(changelog_entry, expected_changelog_entry)

    def test_version_not_found(self):
        import changelog

        with self.assertRaises(RuntimeError) as cm:
            changelog.read_changelogs_news_md(self.news_md, "1.1")
        self.assertEqual(
            str(cm.exception),
            'cannot find expected version "1.1" in first header, found "New in snapd 2.60.3:"',
        )

    def test_deb_email_happy(self):
        import changelog

        original = os.environ.get("DEBEMAIL")
        os.environ["DEBEMAIL"] = "FirstName LastName <firstname.lastname@canonical.com>"
        try:
            changelog.validate_env_deb_email()
        finally:
            if original:
                os.environ["DEBEMAIL"] = original

    def test_deb_email_errors(self):
        import changelog

        original = os.environ.get("DEBEMAIL")
        os.environ["DEBEMAIL"] = ""
        with self.assertRaises(RuntimeError) as e:
            changelog.validate_env_deb_email()
        self.assertEqual(
            str(e.exception),
            'cannot find environment variable "DEBEMAIL", please provide DEBEMAIL="FirstName LastName <valid-email-address>"',
        )
        os.environ["DEBEMAIL"] = "FirstName LastName <firstname.lastname.com>"
        with self.assertRaises(RuntimeError) as e:
            changelog.validate_env_deb_email()
        self.assertEqual(
            str(e.exception),
            'environment variable "DEBEMAIL" uses incorrect format, expecting DEBEMAIL="FirstName LastName <valid-email-address>"',
        )
        if original:
            os.environ["DEBEMAIL"] = original



#!/usr/bin/env python3

import argparse
import logging
import yaml
import sys


def parse_arguments():
    parser = argparse.ArgumentParser(description="ensure pc gadget has ubuntu-save")
    parser.add_argument(
        "gadgetyaml", type=argparse.FileType("r"), help="path to gadget.yaml input file"
    )
    parser.add_argument('--add', help='add ubuntu-save if absent',
                        action='store_true')
    parser.add_argument('--remove', help='remove ubuntu-save if present',
                        action='store_true')
    return parser.parse_args()


def main(opts):
    gadget_yaml = yaml.safe_load(opts.gadgetyaml)

    structs = gadget_yaml["volumes"]["pc"]["structure"]
    save_idx = -1
    if opts.add:
        for idx, s in enumerate(structs):
            role = s.get("role", "")
            if role == "system-save":
                logging.info("system-save structure already present")
                # already has ubuntu-save
                return
            if role == "system-data":
                # ubuntu-save precedes ubuntu-data
                save_idx = idx
                break
        if save_idx == -1:
            raise RuntimeError("cannot find a suitable place to insert ubuntu-save")

        ubuntu_save = {
            "name": "ubuntu-save",
            "role": "system-save",
            # TODO:UC20: update when pc-amd64-gadget changes
            "size": "16M",
            "filesystem": "ext4",
            "type": "83,0FC63DAF-8483-4772-8E79-3D69D8477DE4",
        }
        structs.insert(save_idx, ubuntu_save)
    elif opts.remove:
        for idx, s in enumerate(structs):
            role = s.get("role", "")
            if role == "system-save":
                save_idx = idx
                break
        if save_idx == -1:
                logging.info("system-save structure already absent")
                return
        del structs[save_idx]

    yaml.dump(gadget_yaml, stream=sys.stdout, indent=2, default_flow_style=False)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    opts = parse_arguments()
    if opts.add and opts.remove:
        raise RuntimeError('--add and --remove cannot be used together')
    elif not opts.add and not opts.remove:
        opts.add = True

    main(opts)



#!/usr/bin/env python3

import argparse
from collections import defaultdict
import json
from typing import Any, TextIO
import sys

from features import Cmd, Endpoint, Interface, Task, Change, Ensure, CmdLogLine, EndpointLogLine, InterfaceLogLine, EnsureLogLine, ChangeLogLine, TaskLogLine
from state import State, NotInStateError


def _check_msg(json_entry: dict[str, Any], msg: str) -> bool:
    return 'msg' in json_entry and json_entry['msg'] == msg


def _remove_duplicate_features(key: str, dictionary: dict[str, Any]):
    if key in dictionary:
        l = dictionary[key]
        dictionary[key] = [i for n, i in enumerate(l) if i not in l[n + 1:]]


class CmdFeature:
    name = 'cmd'
    parent = 'cmds'
    msg = 'command-execution'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], _):
        feature_dict[CmdFeature.parent].append(
            Cmd(cmd=json_entry[CmdLogLine.cmd]))

    @staticmethod
    def cleanup_dict(_):
        pass


class EndpointFeature:
    name = 'endpoint'
    parent = 'endpoints'
    msg = 'endpoint'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], _):
        if EndpointLogLine.action in json_entry:
            entry = Endpoint(method=json_entry[EndpointLogLine.method],
                             path=json_entry[EndpointLogLine.path],
                             action=json_entry[EndpointLogLine.action])
        else:
            entry = Endpoint(
                method=json_entry[EndpointLogLine.method], path=json_entry[EndpointLogLine.path])
        feature_dict[EndpointFeature.parent].append(entry)

    @staticmethod
    def cleanup_dict(_):
        pass


class InterfaceFeature:
    name = 'interface'
    parent = 'interfaces'
    msg = 'interface-connection'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], _):
        feature_dict[InterfaceFeature.parent].append(Interface(
            name=json_entry[InterfaceLogLine.interface],
            plug_snap_type=json_entry[InterfaceLogLine.plug],
            slot_snap_type=json_entry[InterfaceLogLine.slot]))

    @staticmethod
    def cleanup_dict(_):
        pass


class EnsureFeature:
    name = 'ensure'
    parent = 'ensures'
    msg = 'ensure'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], _):
        feature_dict[EnsureFeature.parent].append(
                Ensure(manager=json_entry[EnsureLogLine.manager], function=json_entry[EnsureLogLine.func]))

    @staticmethod
    def cleanup_dict(_):
        pass


class ChangeFeature:
    name = 'change'
    parent = 'changes'
    msg = 'new-change'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], state: State):
        snap_types = ['NOT FOUND']
        if state:
            try:
                snap_types = list(state.get_snap_types_from_change_id(
                    json_entry[ChangeLogLine.id]))
            except NotInStateError:
                pass
        feature_dict[ChangeFeature.parent].append(
            Change(kind=json_entry[ChangeLogLine.kind], snap_types=snap_types))

    @staticmethod
    def cleanup_dict(_):
        pass


class TaskFeature:
    name = 'task'
    parent = 'tasks'
    msg = 'task-status-change'

    @staticmethod
    def handle_feature(feature_dict: dict[str, list[Any]], json_entry: dict[str, Any], state: State):
        for entry in feature_dict[TaskFeature.parent]:
            if json_entry[TaskLogLine.id] == entry["id"]:
                entry['last_status'] = json_entry[TaskLogLine.status]
                return
        snap_types = ['NOT FOUND']
        if state:
            try:
                snap_types = state.get_snap_types_from_task_id(
                    json_entry[TaskLogLine.id])
            except NotInStateError:
                pass
        feature_dict[TaskFeature.parent].append(
            Task(id=json_entry[TaskLogLine.id],
                 kind=json_entry[TaskLogLine.task_name],
                 last_status=json_entry[TaskLogLine.status],
                 snap_types=snap_types))

    @staticmethod
    def cleanup_dict(feature_dict: dict[str, list[Any]]):
        if TaskFeature.parent in feature_dict:
            for entry in feature_dict[TaskFeature.parent]:
                del entry['id']


FEATURE_LIST = [CmdFeature, EndpointFeature, InterfaceFeature,
                EnsureFeature, ChangeFeature, TaskFeature]


def get_feature_dictionary(log_file: TextIO, feature_list: list[str], state: State):
    '''
    Extracts features from the journal entries and places them in a dictionary.

    :param log_file: iterator of journal entries
    :param feature_list: list of feature names to extract
    :param state: dictionary of a state.json
    :return: dictionary of features
    :raises: ValueError if an invalid feature name is provided
    :raises: RuntimeError if a line could not be parsed as json
    '''

    feature_dict = defaultdict(list)
    feature_classes = [cls for cls in FEATURE_LIST
                       if cls.name in feature_list]
    if len(feature_classes) != len(feature_list):
        raise ValueError(
            "Error: Invalid feature name in feature list {}".format(feature_list))

    for line in log_file:
        try:
            line_json = json.loads(line)
            for feature_class in feature_classes:
                if _check_msg(line_json, feature_class.msg):
                    try:
                        feature_class.handle_feature(
                            feature_dict, line_json, state)
                    except Exception as e:
                        raise RuntimeError("Encountered error during {} feature processing for {}: {}".format(
                            feature_class.name, line_json, e))
        except json.JSONDecodeError:
            raise RuntimeError("Could not parse line as json: {}".format(line))

    for feature_class in feature_classes:
        feature_class.cleanup_dict(feature_dict)
        _remove_duplicate_features(feature_class.parent, feature_dict)

    return feature_dict


def main():
    parser = argparse.ArgumentParser(
        description="""Given a set of features with journal entries, each in json format, and a 
        state.json, this script will search the text file and extract the features. Those 
        features will be saved in a dictionary and written to the indicated file in output.""")
    parser.add_argument('-o', '--output', help='Output file', required=True)
    parser.add_argument(
        '-f', '--feature', help='Features to extract from journal {cmd, task, change, ensure, endpoint, interface}; '
        'can be repeated multiple times', nargs='+')
    parser.add_argument(
        '-j', '--journal', help='Text file containing journal entries', required=True, type=argparse.FileType('r'))
    parser.add_argument(
        '-s', '--state', help='state.json', required=False, type=argparse.FileType('r'))
    args = parser.parse_args()

    try:
        state = None
        if args.state:
            state = State(json.load(args.state))
        feature_dictionary = get_feature_dictionary(
            args.journal, args.feature, state)
        json.dump(feature_dictionary, open(args.output, "w"))
    except json.JSONDecodeError:
        raise RuntimeError("The state.json is not valid json")


if __name__ == "__main__":
    main()




from typing import Any

CHANGES = 'changes'
TASKS = 'tasks'
SNAP = 'snap'
SNAPS = 'snaps'
TYPE = 'type'
NAME = 'name'
PLUG = 'plug'
SLOT = 'slot'
SNAP_NAME = 'snap-name'
SNAP_TYPE = 'snap-type'
SNAP_SETUP = 'snap-setup'
SNAP_SETUP_TASK = 'snap-setup-task'
SNAP_SETUP_TASKS = 'snap-setup-tasks'
HOOK_SETUP = 'hook-setup'
SIDE_INFO = 'side-info'
DATA = 'data'
SERVICE_ACTION = 'service-action'
QUOTA_CONTROL_ACTIONS = 'quota-control-actions'
RECOVERY_SYSTEM_SETUP_TASK = 'recovery-system-setup-task'
RECOVERY_SYSTEM_SETUP = 'recovery-system-setup'
SNAPSHOT_SETUP = 'snapshot-setup'


TASKS_IGNORE_SNAP_TYPE = {
    # These tasks are not associated with a particular snap
    'clear-confdb-tx-on-error',
    'commit-confdb-tx',
    'clear-confdb-tx',
    'load-confdb-change',
    'update-gadget-cmdline',
    'create-recovery-system',
    'finalize-recovery-system',
    'remove-recovery-system',
    'check-rerefresh',
    'exec-command',
    'request-serial',
    'enforce-validation-sets',
    'mark-seeded',
    'hotplug-add-slot',
    'hotplug-connect',
    'hotplug-disconnect',
    'hotplug-remove-slot',

    # These two are only associated with app/gadget types
    # of snaps and so their snap types can be ignored
    'service-control',
    'quota-control'
}


def _keys_in_dict(dictionary: dict[str, Any], *args) -> bool:
    '''
    Checks if the keys passed as args are in the dictionary,
    each nested inside the other. The first arg is the outermost
    key. So _keys_in_dict(d, 'first', 'second') will check if
    'first' in d and 'second' in d['first']

    :return: True if all keys are present, nested one inside the other.
    '''
    current_entry = dictionary
    for arg in args:
        if arg not in current_entry:
            return False
        current_entry = current_entry[arg]
    return True


class NotInStateError(Exception):
    pass


class State:

    def __init__(self, state_json: dict[str, Any]):
        self.state = state_json

    def get_change(self, id: str) -> dict[str, Any]:
        try:
            return self.state[CHANGES][id]
        except KeyError:
            raise NotInStateError(
                'change {} not found in state.json'.format(id))

    def get_task(self, id: str) -> dict[str, Any]:
        try:
            return self.state[TASKS][id]
        except KeyError:
            raise NotInStateError('task {} not found in state.json'.format(id))

    def get_snap_type(self, snap_name: str) -> str:
        '''
        Given a snap name, returns the type of snap. If the snap name is not present,
        returns the string "NOT_FOUND: " followed by the snap name
        '''
        if _keys_in_dict(self.state, DATA, SNAPS, snap_name, TYPE):
            return self.state[DATA][SNAPS][snap_name][TYPE]
        for task in self.state[TASKS].values():
            if _keys_in_dict(task, DATA, SNAP_SETUP, SIDE_INFO, NAME) \
                    and task[DATA][SNAP_SETUP][SIDE_INFO][NAME] == snap_name \
                    and _keys_in_dict(task, DATA, SNAP_SETUP, TYPE):
                return task[DATA][SNAP_SETUP][TYPE]

        return "NOT_FOUND: {}".format(snap_name)

    def get_snap_types_from_task_id(self, id: str) -> list[str]:
        '''
        Retrieves the type of snap associated with the task with the given id.
        If the task kind is present in the exception list or if the task does
        not have a data section, then returns an empty set.

        :raises KeyError: when the snap type was not found yet the kind of task is not in the exception list
        '''

        task = self.get_task(id)
        if task['kind'] in TASKS_IGNORE_SNAP_TYPE:
            return []
        if DATA not in task:
            return []
        data = task[DATA]

        if _keys_in_dict(data, SNAP_TYPE):
            return [data[SNAP_TYPE]]

        elif _keys_in_dict(data, SNAP_SETUP, TYPE):
            return [data[SNAP_SETUP][TYPE]]

        elif _keys_in_dict(data, SNAP_SETUP, SIDE_INFO, NAME):
            return [self.get_snap_type(data[SNAP_SETUP][SIDE_INFO][NAME])]

        elif _keys_in_dict(data, SNAP_SETUP_TASK):
            return self.get_snap_types_from_task_id(data[SNAP_SETUP_TASK])

        elif _keys_in_dict(data, HOOK_SETUP, SNAP):
            return [self.get_snap_type(data[HOOK_SETUP][SNAP])]

        elif _keys_in_dict(data, PLUG, SNAP) and _keys_in_dict(data, SLOT, SNAP):
            return [self.get_snap_type(data[PLUG][SNAP]), self.get_snap_type(data[SLOT][SNAP])]

        elif _keys_in_dict(data, SNAPS):  # ex: conditional-auto-refresh
            return list({snap_data[TYPE] for snap_data in data[SNAPS].values()})

        elif _keys_in_dict(data, SNAPSHOT_SETUP, SNAP):
            return [self.get_snap_type(data[SNAPSHOT_SETUP][SNAP])]

        elif _keys_in_dict(data, RECOVERY_SYSTEM_SETUP_TASK):
            return self.get_snap_types_from_task_id(data[RECOVERY_SYSTEM_SETUP_TASK])

        elif _keys_in_dict(data, RECOVERY_SYSTEM_SETUP, SNAP_SETUP_TASKS):
            return list({snap_type for setup_task in data[RECOVERY_SYSTEM_SETUP][SNAP_SETUP_TASKS] for snap_type in self.get_snap_types_from_task_id(setup_task)})

        raise NotInStateError(
            'cannot find snap type for task id {} in task {}'.format(id, task))

    def get_snap_types_from_change_id(self, id: str) -> set[str]:
        change = self.get_change(id)
        if 'task-ids' not in change:
            return set()
        tasks = change['task-ids']
        snap_types = set()
        for task in tasks:
            snap_types.update(self.get_snap_types_from_task_id(task))
        return snap_types



# SPDX-FileCopyrightText: 2025 Canonical Ltd
# SPDX-License-Identifier: GPL-3.0-only

import yaml
import sys
import os
from collections import defaultdict

def find_refs(data):
    """Recursively finds all '$ref' values in the data."""
    refs = set()
    if isinstance(data, dict):
        for k, v in data.items():
            if k == '$ref' and isinstance(v, str):
                refs.add(v)
            else:
                refs.update(find_refs(v))
    elif isinstance(data, list):
        for item in data:
            refs.update(find_refs(item))
    return refs

def generate_dot_file(output_filename, source_nodes, endpoint_label, target_groups, all_edges, dark_mode=False):
    """
    Generates a DOT graph file combining multiple target types with a dark mode option.
    """
    # --- Define color palettes ---
    if dark_mode:
        palette = {
            "bgcolor": "#202020",
            "fontcolor": "#d0d0d0",
            "cluster_color": "#CDCDCD",
            "default_edge": "#CDCDCD",
            "endpoints": "#606060",
            "schemas": "#F99B11",
            "responses": "#24598F",
            "security": "#C7162B",
            "schemas_edge": "#F99B11",
            "responses_edge": "#24598F",
            "security_edge": "#C7162B"
        }
    else:
        palette = {
            "bgcolor": "#f8f8f8",
            "fontcolor": "#000000",
            "cluster_color": "#303030",
            "default_edge": "#303030",
            "endpoints": "#606060",
            "schemas": "#F99B11",
            "responses": "#24598F",
            "security": "#C7162B",
            "schemas_edge": "#F99B11",
            "responses_edge": "#24598F",
            "security_edge": "#C7162B"
        }

    with open(output_filename, 'w') as f:
        f.write('digraph OpenAPI_Dependencies {\n')
        f.write(f'  bgcolor="{palette["bgcolor"]}";\n')
        f.write(f'  fontcolor="{palette["fontcolor"]}";\n')
        f.write('  rankdir="TB";\n')
        f.write('  splines="curved";\n')
        f.write('  compound=true;\n')
        f.write('  concentrate=true;\n')
        f.write('  graph [nodesep=0.5, ranksep=2.5];\n')
        f.write(f'  node [shape=box, style="rounded,filled", fontcolor="{palette["fontcolor"]}"];\n')

        # Endpoints Subgraph
        f.write('  subgraph cluster_endpoints {\n')
        f.write(f'    label = "{endpoint_label}";\n')
        f.write(f'    fontcolor = "{palette["cluster_color"]}";\n')
        f.write(f'    color = "{palette["cluster_color"]}";\n')
        f.write('    margin = 10;\n')
        f.write(f'    node [fillcolor="{palette["endpoints"]}"];\n')
        for node in sorted(list(source_nodes)):
            f.write(f'    "{node}";\n')
        f.write('  }\n\n')

        # Target Subgraphs
        for target_label, target_nodes in target_groups.items():
            if target_nodes:
                f.write(f'  subgraph cluster_{target_label.lower()} {{\n')
                f.write(f'    label = "{target_label}";\n')
                f.write(f'    fontcolor = "{palette["cluster_color"]}";\n')
                f.write(f'    color = "{palette["cluster_color"]}";\n')
                f.write('    margin = 10;\n')
                f.write(f'    node [fillcolor="{palette.get(target_label.lower(), "#ffffff")}"];\n')
                for node in sorted(list(target_nodes)):
                    f.write(f'    "{node}";\n')
                f.write('  }\n\n')

        # Logic to color edges based on their target type
        # Create a reverse mapping from a node name to its type (e.g., "User" -> "schemas")
        node_to_type = {}
        for target_label, target_nodes in target_groups.items():
            for node in target_nodes:
                node_to_type[node] = target_label.lower()

        f.write('  // Edges (Dependencies)\n')
        for source, target in sorted(list(all_edges)):
            target_type = node_to_type.get(target, "")
            edge_color = palette.get(f"{target_type}_edge", palette["default_edge"])
            f.write(f'  "{source}" -> "{target}" [color="{edge_color}"];\n')
        
        f.write('}\n')
    print(f"Successfully generated DOT file at '{output_filename}'")

def main(dark_mode=False, max_edges=None):
    """Main function to parse the spec and generate graphs."""
    # The script is now designed to run from the project root.
    # It expects the bundled OpenAPI spec to be in the same directory.
    spec_file = 'openapi-bundled.yaml'
    project_root = os.getcwd()
    
    # Determine output directory based on dark mode
    theme = 'dark' if dark_mode else 'light'
    output_dir = os.path.join(project_root, 'visuals', theme)
    
    os.makedirs(output_dir, exist_ok=True)
    print(f"Output directory '{output_dir}' is ready.")

    try:
        with open(spec_file, 'r') as f:
            spec = yaml.safe_load(f)
    except FileNotFoundError:
        print(f"Error: The file '{spec_file}' was not found.")
        print("Please make sure to run 'make bundle' to generate the bundled OpenAPI specification.")
        sys.exit(1)
    except yaml.YAMLError as e:
        print(f"Error parsing YAML file: {e}")
        sys.exit(1)

    valid_http_methods = {'get', 'put', 'post', 'delete', 'options', 'head', 'patch', 'trace'}

    paths_by_tag = defaultdict(dict)
    for path, path_item in spec.get('paths', {}).items():
        for method in set(path_item.keys()) & valid_http_methods:
            operation = path_item[method]
            if not isinstance(operation, dict):
                continue

            tags = operation.get('tags', ['untagged'])
            for tag in tags:
                if path not in paths_by_tag[tag]:
                    paths_by_tag[tag][path] = {}
                paths_by_tag[tag][path][method] = operation

    for tag, paths in paths_by_tag.items():
        print(f"\n--- Processing tag: {tag} ---")
        
        source_nodes = set()
        edges_to_schemas = set()
        edges_to_responses = set()
        edges_to_security = set()

        for path, path_item in paths.items():
            for method, operation in path_item.items():
                path_node_name = f"{method.upper()} {path}"
                source_nodes.add(path_node_name)
            
                if 'security' in operation and operation['security'] is not None:
                    for security_req in operation['security']:
                        for scheme_name in security_req.keys():
                            edges_to_security.add((path_node_name, scheme_name))

                all_refs = find_refs(operation)
                for ref in all_refs:
                    target_name = ref.split('/')[-1]
                    if '#/components/schemas/' in ref:
                        edges_to_schemas.add((path_node_name, target_name))
                    elif '#/components/responses/' in ref:
                        edges_to_responses.add((path_node_name, target_name))

        endpoint_label = f"Endpoints - {tag}"

        all_edges = edges_to_schemas.union(edges_to_responses).union(edges_to_security)
        
        if max_edges is not None and len(all_edges) > max_edges:
            print(f"Skipping DOT file generation for tag '{tag}' because it has {len(all_edges)} edges, which exceeds the max of {max_edges}.")
            continue

        if all_edges:
            schema_nodes = {edge[1] for edge in edges_to_schemas}
            response_nodes = {edge[1] for edge in edges_to_responses}
            security_nodes = {edge[1] for edge in edges_to_security}
            
            target_groups = {
                "Schemas": schema_nodes,
                "Responses": response_nodes,
                "Security": security_nodes
            }
            
            output_path = os.path.join(output_dir, f'{tag.replace(" ", "_")}_dependencies.dot')
            generate_dot_file(output_path, source_nodes, endpoint_label, target_groups, all_edges, dark_mode)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Generate DOT dependency graphs from an OpenAPI specification.')
    parser.add_argument('--dark', action='store_true', help='Enable dark mode for the generated graphs.')
    parser.add_argument('--max-edges', type=int, help='Skip graph generation if the number of edges exceeds this limit.')
    args = parser.parse_args()

    main(dark_mode=args.dark, max_edges=args.max_edges)



#!/usr/bin/python3

# This script checks that the committer and author email addresses of commits
# are valid - this is important because we need to be able to import the git
# repo to launchpad and launchpad does not like git commits that are formatted
# badly.
# This can be run either on a merge commit in which case the commits it
# evaluates is just limited to the "other" branch (the one being merged into the
# destination) like we do in CI workflows. Or it can be run on a normal commit
# like a developer would do locally in which case this ends up checking all
# commits locally - some day that may grow to be unwieldy but for now seems ok.

import os
import subprocess
from email.utils import parseaddr


def get_commit_range():
    # For CI, the head revision is a synthesised merge commit,
    # merging the proposed branch into the destination branch.
    # So the first parent is our destination, and the second is
    # our proposal.
    lines = subprocess.check_output(
        ["git", "cat-file", "-p", "@"], text=True
    ).splitlines()
    parents = [
        line[len("parent ") :].strip() for line in lines if line.startswith("parent ")
    ]
    if len(parents) == 1:
        # not a merge commit, so return nothing to use default git log behavior
        # and check all commits
        return ""
    elif len(parents) == 2:
        # merge commit so use "foo..bar" syntax to just check the proposed
        # commits
        dest, proposed = parents
        return "{}..{}".format(dest, proposed)
    else:
        raise RuntimeError("expected two parents, but got {}".format(parents))


def exception_invalid(maybeaddr: str) -> bool:
    # from commit d935048f4ee5f9a8cde640a760bf1fdb9409de7e
    # dependabot[bot] <49699333+dependabot[bot]@users.noreply.github.com>
    if "+dependabot[bot]@users.noreply.github.com" in maybeaddr:
        return True
    return False


if __name__ == "__main__":
    if not os.path.exists('.git'):
        exit(0)
    commitrange = get_commit_range()
    args = ["git", "log", "--format=format:%h,%ce%n%h,%ae"]
    if commitrange != "":
        args.append(commitrange)
    for line in subprocess.check_output(args, text=True).split("\n"):
        parsed = line.split(",", 1)
        commithash = parsed[0]
        potentialemail = parsed[1]
        if potentialemail == "":
            continue
        name, addr = parseaddr(potentialemail)
        if addr == "" and not exception_invalid(potentialemail):
            print(
                "Found invalid email %s for commmit %s" % (potentialemail, commithash)
            )
            exit(1)
        if not addr.isascii():
            print(
                "Found invalid non-ascii email %s for commmit %s"
                % (potentialemail, commithash)
            )
            exit(1)



#!/usr/bin/env python3

import os
import sys
import yaml


if len(sys.argv) < 2:
    print('2 arguments are required: gadget.yaml and size')
    sys.exit(1)

gadget = sys.argv[1]
size = sys.argv[2]

with open(gadget, 'r') as f:
    data = yaml.safe_load(f)

for entry in data['volumes']['pc']['structure']:
    if entry.get('role') == 'system-seed':
        entry['size'] = size

with open(sys.argv[1], 'w') as f:
    yaml.dump(data, f)



import unittest
from io import StringIO

from common import LockOpTrace, get_next_match

from filter import LockOp, LocksGroup, GroupTimes, LockTraceManager, LocksFileReader


class TestLockOp(unittest.TestCase):
    def setUp(self):
        self.lines = ["header held: 10 ms wait 5 ms\n", "line 1\n", "line 2\n"]
        self.lock_op = LockOp(self.lines)

    def test_get_held_time(self):
        self.assertEqual(self.lock_op.get_held_time(), 10)

    def test_get_wait_time(self):
        self.assertEqual(self.lock_op.get_wait_time(), 5)

    def test_get_trace(self):
        expected_trace = LockOpTrace(["line 1\n", "line 2\n"])
        self.assertEqual(self.lock_op.get_trace(), expected_trace)


class TestLocksGroup(unittest.TestCase):
    def setUp(self):
        self.lines = [
            "###START: Test Group\n",
            "### Lock 1 header: held: 10 ms wait 5 ms\n",
            "line 1\n",
            "line 2\n",
            "### Lock 1 header: held: 20 ms wait 10 ms\n",
            "line 3\n",
        ]
        self.group = LocksGroup(self.lines)

    def test_get_name(self):
        self.assertEqual(self.group.get_name(), "###START: Test Group\n")

    def test_get_locks(self):
        self.assertEqual(len(self.group.get_locks()), 2)

    def test_get_traces(self):
        self.assertEqual(len(self.group.get_traces()), 2)

    def test_get_lock_held_time(self):
        trace = self.group.get_traces()[0]
        self.assertEqual(self.group.get_lock_held_time(trace), 10)

    def test_get_lock_wait_time(self):
        trace = self.group.get_traces()[0]
        self.assertEqual(self.group.get_lock_wait_time(trace), 5)


class TestGroupTimes(unittest.TestCase):
    def setUp(self):
        self.group_time = GroupTimes("Test Group", 10, 5)

    def test_get_group_name(self):
        self.assertEqual(self.group_time.get_group_name(), "Test Group")

    def test_get_held_time(self):
        self.assertEqual(self.group_time.get_held_time(), 10)

    def test_get_wait_time(self):
        self.assertEqual(self.group_time.get_wait_time(), 5)


class TestLockTraceManager(unittest.TestCase):
    def setUp(self):
        trace = LockOpTrace(["line 1\n", "line 2\n"])
        group_time = GroupTimes("Test Group", 10, 5)
        self.manager = LockTraceManager({trace: [group_time]})

    def test_filter(self):
        self.manager.filter(held_time=5, wait_time=3)
        self.assertEqual(len(self.manager.traces), 1)
        self.manager.filter(held_time=15, wait_time=3)
        self.assertEqual(len(self.manager.traces), 0)

    def test_match(self):
        self.manager.match(["line 1"])
        self.assertEqual(len(self.manager.traces), 1)
        self.manager.match(["line 3"])
        self.assertEqual(len(self.manager.traces), 0)


class TestLocksFileReader(unittest.TestCase):
    def setUp(self):
        self.file_content = (
            "###START: SNAPD PROJECT\n"
            "### Lock 1 header: held: 0 ms wait 50 ms\n"
            "line 1\n"
            "### Lock 2 header: held: 100 ms wait 0 ms\n"
            "line 2\n"
            "###START: Test 1\n"
            "### Lock 3 header: held: 10 ms wait 5 ms\n"
            "line 3\n"
            "line 4\n"
            "line 5\n"
        )
        self.file = StringIO(self.file_content)
        self.reader = LocksFileReader(self.file)

    def test_get_test(self):
        test_output = self.reader.get_test("Test 1")
        self.assertIn("###START: Test 1\n", test_output)

    def test_get_traces_times(self):
        traces_times = self.reader.get_traces_times()
        self.assertEqual(len(traces_times), 3)


class TestGetNextMatch(unittest.TestCase):
    def test_get_next_match(self):
        lines = ["line 1\n", "line 2\n", "###START: Test\n", "line 3\n"]
        result = get_next_match(lines, 0, "###START:")
        self.assertEqual(result, 2)
        result = get_next_match(lines, 2, "###START:")
        self.assertEqual(result, -1)


if __name__ == "__main__":
    unittest.main()



import argparse
import json
from importlib.util import spec_from_loader, module_from_spec
from importlib.machinery import SourceFileLoader
from io import StringIO
import os
from typing import Any, Tuple, TypedDict, Literal, Union
import unittest
from unittest.mock import Mock, patch

# Since log-analyzer has a hyphen and is missing the .py extension,
# we need to do some extra work to import the module to test
dir_path = os.path.dirname(os.path.realpath(__file__))
spec = spec_from_loader(
    "log-analyzer",
    SourceFileLoader("log-analyzer", os.path.join(dir_path, "log-analyzer")),
)
if spec is None:
    raise RuntimeError("cannot get log-analyzer spec")
if spec.loader is None:
    raise RuntimeError("cannot get log-analyzer spec loader")
log_analyzer = module_from_spec(spec)
if log_analyzer is None:
    raise RuntimeError("cannot get log-analyzer spec")
spec.loader.exec_module(log_analyzer)


class SpreadLog_TypePhase(TypedDict):
    type: Literal["phase"]
    task: str
    verb: str


class SpreadLogDetail(TypedDict):
    lines: list[str]


class SpreadLog_TypeResult(TypedDict):
    type: Literal["result"]
    result_type: str
    level: str
    stage: str
    detail: SpreadLogDetail


SpreadLog = Union[SpreadLog_TypePhase, SpreadLog_TypeResult]


def create_data(
    num_executed_no_fail: int,
    num_fail_execution: int,
    num_fail_restore: int,
    num_fail_prepare: int,
    num_not_executed: int,
) -> Tuple[set[str], list[SpreadLog]]:
    # The order will be:
    #   1. tasks that executed and didn't fail
    #   2. tasks that failed during execution
    #   3. tasks that failed during restore
    #   4. tasks that failed during prepare
    #   5. tasks that were not executed at all

    exec_param = [
        "test_" + str(i)
        for i in range(
            num_executed_no_fail
            + num_fail_execution
            + num_fail_prepare
            + num_fail_restore
            + num_not_executed
        )
    ]

    # The tasks that executed are those that didn't fail plus those that failed during execution or restore
    spread_logs: list[SpreadLog] = [
        SpreadLog_TypePhase(
            {"type": "phase", "verb": "Executing", "task": param})
        for param in exec_param[
            : num_executed_no_fail + num_fail_execution + num_fail_restore
        ]
    ]

    begin = num_executed_no_fail
    end = num_executed_no_fail + num_fail_execution
    # The tasks that failed are those that failed during execution, not during restore or prepare
    spread_logs.append(
        SpreadLog_TypeResult(
            {
                "type": "result",
                "result_type": "Failed",
                "level": "tasks",
                "stage": "",
                "detail": {
                    "lines": ["- %s\n" % param for param in exec_param[begin:end]]
                },
            }
        )
    )

    begin = num_executed_no_fail + num_fail_execution
    end = num_executed_no_fail + num_fail_execution + num_fail_restore
    # Tasks that failed during the restore phase
    spread_logs.append(
        SpreadLog_TypeResult(
            {
                "type": "result",
                "result_type": "Failed",
                "level": "task",
                "stage": "restore",
                "detail": {
                    "lines": ["- %s\n" % param for param in exec_param[begin:end]]
                },
            }
        )
    )

    begin = num_executed_no_fail + num_fail_execution + num_fail_restore
    end = (
        num_executed_no_fail + num_fail_execution + num_fail_restore + num_fail_prepare
    )
    # Tasks that failed during the prepare phase
    spread_logs.append(
        SpreadLog_TypeResult(
            {
                "type": "result",
                "result_type": "Failed",
                "level": "task",
                "stage": "prepare",
                "detail": {
                    "lines": ["- %s\n" % param for param in exec_param[begin:end]]
                },
            }
        )
    )

    return set(exec_param), spread_logs


class TestLogAnalyzer(unittest.TestCase):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(TestLogAnalyzer, self).__init__(*args, **kwargs)

        self.filtered_exec_param_mixed, self.spread_logs_mixed = create_data(
            num_executed_no_fail=10,
            num_fail_execution=10,
            num_fail_restore=10,
            num_fail_prepare=10,
            num_not_executed=10,
        )
        self.exec_param_mixed = ["tests/...", "other-tests/..."]

        self.filtered_exec_param_no_failed, self.spread_logs_no_failed = create_data(
            num_executed_no_fail=10,
            num_fail_execution=0,
            num_fail_restore=0,
            num_fail_prepare=0,
            num_not_executed=0,
        )
        self.exec_param_no_failed = ["tests/...", "other-tests/..."]

        self.filtered_exec_param_no_exec, self.spread_logs_no_exec = create_data(
            num_executed_no_fail=0,
            num_fail_execution=0,
            num_fail_restore=0,
            num_fail_prepare=10,
            num_not_executed=10,
        )
        self.exec_param_no_exec = ["tests/...", "other-tests/..."]

        (
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        ) = create_data(
            num_executed_no_fail=10,
            num_fail_execution=0,
            num_fail_restore=0,
            num_fail_prepare=0,
            num_not_executed=10,
        )
        self.exec_param_mix_success_abort = ["tests/...", "other-tests/..."]

    # The following test group has mixed results with task results
    # of all kinds: successful, failed in all three phases, and not run

    def test_list_executed__mixed(self) -> None:
        actual = log_analyzer.list_executed_tasks(
            self.filtered_exec_param_mixed, self.spread_logs_mixed
        )
        expected = set(["test_" + str(i) for i in range(30)])
        self.assertSetEqual(expected, actual)

    def test_list_failed__mixed(self) -> None:
        actual = log_analyzer.list_failed_tasks(
            self.filtered_exec_param_mixed, self.spread_logs_mixed
        )
        expected = set(["test_" + str(i) for i in range(10, 20)])
        self.assertSetEqual(expected, actual)

    def test_list_successful__mixed(self) -> None:
        actual = log_analyzer.list_successful_tasks(
            self.filtered_exec_param_mixed, self.spread_logs_mixed
        )
        expected = set(["test_" + str(i) for i in range(10)])
        self.assertSetEqual(expected, actual)

    def test_executed_and_failed__mixed(self) -> None:
        actual = log_analyzer.list_executed_and_failed(
            self.filtered_exec_param_mixed, self.spread_logs_mixed
        )
        expected = set(["test_" + str(i) for i in range(10, 30)])
        self.assertSetEqual(expected, actual)

    def test_aborted_tasks__mixed(self) -> None:
        actual = log_analyzer.list_aborted_tasks(
            self.filtered_exec_param_mixed, self.spread_logs_mixed
        )
        expected = set(["test_" + str(i) for i in range(30, 50)])
        self.assertSetEqual(expected, actual)

    def test_reexecute_tasks__mixed(self) -> None:
        actual = log_analyzer.list_rexecute_tasks(
            self.exec_param_mixed,
            self.filtered_exec_param_mixed,
            self.spread_logs_mixed,
        )
        expected = set(["test_" + str(i) for i in range(10, 50)])
        self.assertSetEqual(expected, actual)

    # The following test group has only tasks that were successfully run

    def test_list_executed__no_fail(self) -> None:
        actual = log_analyzer.list_executed_tasks(
            self.filtered_exec_param_no_failed, self.spread_logs_no_failed
        )
        expected = set(["test_" + str(i) for i in range(10)])
        self.assertSetEqual(expected, actual)

    def test_list_failed__no_fail(self) -> None:
        actual = log_analyzer.list_failed_tasks(
            self.filtered_exec_param_no_failed, self.spread_logs_no_failed
        )
        self.assertEqual(0, len(actual))

    def test_list_successful__no_fail(self) -> None:
        actual = log_analyzer.list_successful_tasks(
            self.filtered_exec_param_no_failed, self.spread_logs_no_failed
        )
        expected = set(["test_" + str(i) for i in range(10)])
        self.assertSetEqual(expected, actual)

    def test_executed_and_failed__no_fail(self) -> None:
        actual = log_analyzer.list_executed_and_failed(
            self.filtered_exec_param_no_failed, self.spread_logs_no_failed
        )
        self.assertEqual(0, len(actual))

    def test_aborted_tasks__no_fail(self) -> None:
        actual = log_analyzer.list_aborted_tasks(
            self.filtered_exec_param_no_failed, self.spread_logs_no_failed
        )
        self.assertEqual(0, len(actual))

    def test_reexecute_tasks__no_fail(self) -> None:
        actual = log_analyzer.list_rexecute_tasks(
            self.exec_param_no_failed,
            self.filtered_exec_param_no_failed,
            self.spread_logs_no_failed,
        )
        self.assertEqual(0, len(actual))

    # The following group only has tasks that either failed
    # during the prepare phase or were not run at all

    def test_list_executed__no_exec(self) -> None:
        actual = log_analyzer.list_executed_tasks(
            self.filtered_exec_param_no_exec, self.spread_logs_no_exec
        )
        self.assertEqual(0, len(actual))

    def test_list_failed__no_exec(self) -> None:
        actual = log_analyzer.list_failed_tasks(
            self.filtered_exec_param_no_exec, self.spread_logs_no_exec
        )
        self.assertEqual(0, len(actual))

    def test_list_successful__no_exec(self) -> None:
        actual = log_analyzer.list_successful_tasks(
            self.filtered_exec_param_no_exec, self.spread_logs_no_exec
        )
        self.assertEqual(0, len(actual))

    def test_executed_and_failed__no_exec(self) -> None:
        actual = log_analyzer.list_executed_and_failed(
            self.filtered_exec_param_no_exec, self.spread_logs_no_exec
        )
        self.assertEqual(0, len(actual))

    def test_aborted_tasks__no_exec(self) -> None:
        actual = log_analyzer.list_aborted_tasks(
            self.filtered_exec_param_no_exec, self.spread_logs_no_exec
        )
        expected = set(["test_" + str(i) for i in range(20)])
        self.assertSetEqual(expected, actual)

    def test_reexecute_tasks__no_exec(self) -> None:
        actual = log_analyzer.list_rexecute_tasks(
            self.exec_param_no_exec,
            self.filtered_exec_param_no_exec,
            self.spread_logs_no_exec,
        )
        self.assertSetEqual(set(self.exec_param_no_exec), actual)

    # The following test group has tasks that either
    # were successful or did not run at all

    def test_list_executed__mix_success_abort(self) -> None:
        actual = log_analyzer.list_executed_tasks(
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        expected = set(["test_" + str(i) for i in range(10)])
        self.assertSetEqual(expected, actual)

    def test_list_failed__mix_success_abort(self) -> None:
        actual = log_analyzer.list_failed_tasks(
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        self.assertEqual(len(actual), 0)

    def test_list_successful__mix_success_abort(self) -> None:
        actual = log_analyzer.list_successful_tasks(
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        expected = set(["test_" + str(i) for i in range(10)])
        self.assertSetEqual(expected, actual)

    def test_executed_and_failed__mix_success_abort(self) -> None:
        actual = log_analyzer.list_executed_and_failed(
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        self.assertEqual(len(actual), 0)

    def test_aborted_tasks__mix_success_abort(self) -> None:
        actual = log_analyzer.list_aborted_tasks(
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        expected = set(["test_" + str(i) for i in range(10, 20)])
        self.assertSetEqual(expected, actual)

    def test_reexecute_tasks__mix_success_abort(self) -> None:
        actual = log_analyzer.list_rexecute_tasks(
            self.exec_param_mix_success_abort,
            self.filtered_exec_param_mix_success_abort,
            self.spread_logs_mix_success_abort,
        )
        expected = set(["test_" + str(i) for i in range(10, 20)])
        self.assertSetEqual(expected, actual)

    # The following test group checks the main function with
    # mixed results (some failures, some aborts, some successes)

    @patch('argparse.ArgumentParser.parse_args')
    def test_list_executed__main(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-executed-tasks',
            exec_params=' '.join(self.exec_param_mixed),
            parsed_log=StringIO(json.dumps(self.spread_logs_mixed))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(["test_" + str(i) for i in range(30)])
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))

    @patch('argparse.ArgumentParser.parse_args')
    def test_list_failed__main(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-failed-tasks',
            exec_params=' '.join(self.exec_param_mixed),
            parsed_log=StringIO(json.dumps(self.spread_logs_mixed))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(["test_" + str(i) for i in range(10, 20)])
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))

    @patch('argparse.ArgumentParser.parse_args')
    def test_list_successful__main(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-successful-tasks',
            exec_params=' '.join(self.exec_param_mixed),
            parsed_log=StringIO(json.dumps(self.spread_logs_mixed))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(["test_" + str(i) for i in range(10)])
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))

    @patch('argparse.ArgumentParser.parse_args')
    def test_aborted_tasks__main(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-aborted-tasks',
            exec_params=' '.join(self.exec_param_mixed),
            parsed_log=StringIO(json.dumps(self.spread_logs_mixed))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(["test_" + str(i) for i in range(30, 50)])
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))

    @patch('argparse.ArgumentParser.parse_args')
    def test_reexecute_tasks__main(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-reexecute-tasks',
            exec_params=' '.join(self.exec_param_mixed),
            parsed_log=StringIO(json.dumps(self.spread_logs_mixed))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(["test_" + str(i) for i in range(10, 50)])
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))

    @patch('argparse.ArgumentParser.parse_args')
    def test_reexecute_tasks__main_no_exec(self, parse_args_mock: Mock) -> None:
        log_analyzer.filter_with_spread = Mock()
        log_analyzer.filter_with_spread.return_value = [
            "test_" + str(i) for i in range(50)]
        parse_args_mock.return_value = argparse.Namespace(
            command='list-reexecute-tasks',
            exec_params=' '.join(self.exec_param_no_exec),
            parsed_log=StringIO(json.dumps(self.spread_logs_no_exec))
        )
        with patch('sys.stdout', new=StringIO()) as stdout_patch:
            log_analyzer.main()
            expected = set(self.exec_param_no_exec)
            self.assertSetEqual(expected, set(stdout_patch.getvalue().split()))


if __name__ == "__main__":
    unittest.main()




'''
Dictionaries to specify structure of feature logs
'''

from enum import Enum
from typing import TypedDict


class Cmd(TypedDict):
    cmd: str


class Endpoint(TypedDict):
    method: str
    path: str
    action: str


class Interface(TypedDict):
    name: str
    plug_snap_type: str
    slot_snap_type: str


class Status(str, Enum):
    done = "Done"
    undone = "Undone"
    error = "Error"

class Task(TypedDict):
    id: str
    kind: str
    snap_types: list[str]
    last_status: str


class Change(TypedDict):
    kind: str
    snap_types: list[str]


class Ensure(TypedDict):
    manager: str
    function: str


class EnvVariables(TypedDict):
    name: str
    value: str

class TaskFeatures(TypedDict):
    suite: str
    task_name: str
    variant: str
    success: bool
    cmds: list[Cmd]
    endpoints: list[Endpoint]
    interfaces: list[Interface]
    tasks: list[Task]
    changes: list[Change]
    ensures: list[Ensure]


class SystemFeatures(TypedDict):
    schema_version: str
    system: str
    scenarios: list[str]
    env_variables: list[EnvVariables]
    tests: list[TaskFeatures]


class CmdLogLine:
    cmd = 'cmd'


class EndpointLogLine:
    method = 'method'
    path = 'path'
    action = 'action'


class InterfaceLogLine:
    interface = 'interface'
    slot = 'slot'
    plug = 'plug'


class EnsureLogLine:
    manager = 'manager'
    func = 'func'


class TaskLogLine:
    task_name = 'task-name'
    id = 'id'
    status = 'status'


class ChangeLogLine:
    kind = 'kind'
    id = 'id'



#!/usr/bin/env python3

# verify-dl verifies all ELF binaries with an interpreter have the
# correct path for its interpreter.

import os
import sys
import logging
import argparse

from elftools.elf.elffile import ELFFile


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="verify requested interpreter of ELF binaries"
    )
    craft_prime = os.environ.get("CRAFT_PRIME")
    parser.add_argument(
        "--prime",
        default=craft_prime,
        required=(craft_prime is None),
        help="snapcraft part priming location (or rootdir of a snap)",
    )
    parser.add_argument("interp", help="Expected interpreter")

    return parser.parse_args()


def main(opts) -> None:
    errors: list[str] = []

    for dirpath, _, filenames in os.walk(opts.prime):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            if os.path.islink(path):
                continue
            with open(path, "rb") as f:
                if f.read(4) != b"\x7fELF":
                    continue
                f.seek(0, 0)

                logging.debug("checking ELF binary: %s", path)

                elf = ELFFile(f)
                for segment in elf.iter_segments():
                    # TODO: use iter_segments(type='PT_INTERP')
                    if (
                        segment["p_type"] == "PT_INTERP"
                        and segment.get_interp_name() != opts.interp
                    ):
                        logging.error(
                            '%s: expected interpreter to be "%s", got "%s"',
                            path,
                            sys.argv[2],
                            segment.get_interp_name(),
                        )
                        errors.append(path)

    if errors:
        badlist = "\n".join(["- " + n for n in errors])
        raise RuntimeError(f"found binaries with incorrect ELF interpreter:\n{badlist}")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main(parse_arguments())



import os
import subprocess
import unittest


def is_24_04():
    with open("/etc/os-release") as inf:
        return "VERSION_ID=\"24.04\"" in inf.read()


class TestFlake8(unittest.TestCase):
    @unittest.skipIf(is_24_04(), "flake8 is broken on 24.04")
    def test_flake8(self):
        p = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
        subprocess.check_call(["flake8", "--ignore=E501", p])



#!/usr/bin/env python3
import dbus
import sys


def run():
    obj = dbus.SystemBus().get_object(
        "org.freedesktop.hostname1", "/org/freedesktop/hostname1"
    )
    print(obj.Introspect(dbus_interface="org.freedesktop.DBus.Introspectable"))


if __name__ == "__main__":
    sys.exit(run())



import argparse
import io
import json
import subprocess


def load_json(disk):
    raw = subprocess.run(['cryptsetup', 'luksDump', '--dump-json-metadata', disk], capture_output=True)
    return json.loads(raw.stdout)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--read',
                        action='store_true')
    parser.add_argument('--dry-run',
                        action='store_true')
    parser.add_argument('disk')
    parser.add_argument('dump')
    args = parser.parse_args()

    if args.read:
        data = load_json(args.disk)
        with open(args.dump, 'w') as f:
            json.dump(data, f)
    else:
        with open(args.dump, 'r') as f:
            backup = json.load(f)
        current = load_json(args.disk)
        for k, _ in current['tokens'].items():
            cmd = ['cryptsetup', 'token', 'remove', '--token-id', k, args.disk]
            if args.dry_run:
                print(' '.join(cmd))
            else:
                subprocess.run(cmd)
        for k, v in backup['tokens'].items():
            token_raw = json.dumps(v)
            cmd = ['cryptsetup', 'token', 'import', '--token-id', k, args.disk]
            if args.dry_run:
                print(' '.join(cmd))
                print(token_raw)
            else:
                p = subprocess.Popen(cmd, stdin=subprocess.PIPE)
                p.communicate(token_raw.encode('utf-8'))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2025 Canonical Ltd
# SPDX-License-Identifier: GPL-3.0-only

"""
This script injects a block of HTML into a file just before a specified closing tag.

It is used to insert custom content, such as CSS or a footer, into the
statically generated Redoc documentation.
"""

import argparse
import sys

def inject_html(file_path: str, html_content: str, html_to_inject: str, target_tag: str) -> str:
    """
    Injects a string of HTML into a file before a specified closing tag.

    Args:
        file_path: The path to the HTML file.
        html_content: The full HTML content of the page.
        html_to_inject: The HTML string to inject.
        target_tag: The closing tag to inject before (e.g., '</head>').
    
    Returns:
        The modified HTML content.
    """
    if target_tag not in html_content:
        print(
            f"Error: Target tag '{target_tag}' not found in '{file_path}'.",
            file=sys.stderr,
        )
        sys.exit(1)

    # Inject the HTML before the target tag
    modified_content = html_content.replace(target_tag, f"{html_to_inject}{target_tag}")
    return modified_content


def set_dark_theme(html_content: str) -> str:
    """
    Sets the dark theme by adding data-theme="dark" to the <body> tag.

    Args:
        html_content: The HTML content.

    Returns:
        The modified HTML content with the dark theme enabled.
    """
    return html_content.replace("<body>", '<body data-theme="dark">')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Inject HTML into a file and set the theme."
    )
    parser.add_argument(
        "file_path",
        help="The path to the HTML file.",
    )
    parser.add_argument(
        "html_file",
        nargs="?",
        default=None,
        help="The path to the file containing the HTML to inject (optional).",
    )
    parser.add_argument(
        "target_tag",
        nargs="?",
        default=None,
        help="The closing tag to inject before (e.g., '</head>') (optional).",
    )

    args = parser.parse_args()

    try:
        with open(args.file_path, "r", encoding="utf-8") as f:
            html_content = f.read()
    except FileNotFoundError:
        print(f"Error: File not found at '{args.file_path}'.", file=sys.stderr)
        sys.exit(1)

    if args.html_file and args.target_tag:
        try:
            with open(args.html_file, "r", encoding="utf-8") as f:
                html_to_inject = f.read()
        except FileNotFoundError:
            print(
                f"Error: HTML file not found at '{args.html_file}'.",
                file=sys.stderr,
            )
            sys.exit(1)
        # Set the dark theme first
        modified_content = set_dark_theme(html_content)
        # Then inject the CSS
        modified_content = inject_html(
            args.file_path, modified_content, html_to_inject, args.target_tag
        )
    else:
        modified_content = html_content

    try:
        with open(args.file_path, "w", encoding="utf-8") as f:
            f.write(modified_content)
        print(f"Successfully processed and wrote to '{args.file_path}'.")
    except Exception as e:
        print(f"An error occurred while writing to file: {e}", file=sys.stderr)
        sys.exit(1)



import yaml
import sys

with open(sys.argv[1], 'r') as f:
    data = yaml.safe_load(f)

for entry in data['volumes']['pc']['structure']:
    if entry.get('role') == 'system-seed':
        entry['role'] = 'system-seed-null'
        entry['name'] = 'EFI System partition'
        # TODO make this realistically smaller?
        entry['size'] = '99M'
    if entry.get('role') == 'system-boot':
        # Such that potentially there is space to later slot-in 1200M
        # large ubuntu-seed partition
        entry['offset'] = '1202M'

with open(sys.argv[1], 'w') as f:
    yaml.dump(data, f)



#!/usr/bin/python3

import argparse
import re


class InvalidPRTitle(Exception):
    def __init__(self, invalid_title):
        self.invalid_title = invalid_title


def check_pr_title(pr_title: str):
    print(pr_title)
    # cover most common cases:
    # package: foo
    # package, otherpackage/subpackage: this is a title
    # tests/regression/lp-12341234: foo
    # [RFC] foo: bar
    if not re.match(r"[a-zA-Z0-9_\-\*/,. \[\](){}]+: .*", pr_title):
        raise InvalidPRTitle(pr_title)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "pr_title", metavar="PR title", help="the github title to check"
    )
    args = parser.parse_args()
    try:
        check_pr_title(args.pr_title)
    except InvalidPRTitle as e:
        print('Invalid PR title: "{}"\n'.format(e.invalid_title))
        print("Please provide a title in the following format:")
        print("module: short description")
        print("E.g.:")
        print("daemon: fix frobinator bug")
        raise SystemExit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Extract tagged code snippets from HACKING.md for testing.

Usage:
    extract-snippets.py MARKDOWN_FILE TAG [TAG...]

Extracts code blocks preceded by <!-- test:TAG --> comments.
Outputs the extracted code to stdout.
"""

import re
import sys
from typing import Dict, List


def extract_snippets(filename: str, tags: List[str]) -> Dict[str, str]:
    """Extract code blocks preceded by <!-- test:TAG --> comments."""
    with open(filename, "r") as f:
        content = f.read()

    snippets: Dict[str, str] = {}

    for tag in tags:
        # Match: <!-- test:TAG -->\n```[optional lang]\nCODE\n```
        # Also match indented code blocks (4 spaces)
        pattern = rf"<!--\s*test:{re.escape(tag)}\s*-->\s*\n```[^\n]*\n(.*?)\n```"
        matches = re.findall(pattern, content, re.DOTALL)

        # Also try to match indented code blocks (HACKING.md uses both styles)
        if not matches:
            # Match: <!-- test:TAG -->\n    code\n    code\n\n
            pattern = rf"<!--\s*test:{re.escape(tag)}\s*-->\s*\n((?:    .*\n)+)"
            matches = re.findall(pattern, content)
            if matches:
                # Remove 4-space indentation
                matches = [
                    "\n".join(
                        line[4:] if line.startswith("    ") else line
                        for line in match.split("\n")
                    )
                    for match in matches
                ]

        if matches:
            snippets[tag] = "\n".join(matches)

    return snippets


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print(
            f"Usage: {sys.argv[0]} MARKDOWN_FILE TAG [TAG...]", file=sys.stderr
        )
        sys.exit(1)

    markdown_file = sys.argv[1]
    tags = sys.argv[2:]

    try:
        snippets = extract_snippets(markdown_file, tags)
    except FileNotFoundError:
        print(f"Error: File '{markdown_file}' not found", file=sys.stderr)
        sys.exit(1)

    for tag in tags:
        if tag in snippets:
            print(snippets[tag])
        else:
            print(
                f"Error: Tag '{tag}' not found in {markdown_file}",
                file=sys.stderr,
            )
            sys.exit(1)



"""
This library provides a set of functions that can be used to
process the spread output.
"""

import re

from enum import Enum


class ListedEnum(Enum):
    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))

class ExecutionPhase(ListedEnum):
    PREPARING = "Preparing"
    EXECUTING = "Executing"
    RESTORING = "Restoring"


class ExecutionInfo(ListedEnum):
    DEBUG = "Debug"
    ERROR = "Error"
    WARNING = "WARNING:"


class ExecutionLevel(ListedEnum):
    TASK = "task"
    SUITE = "suite"
    PROJECT = "project"


class GeneralAction(ListedEnum):
    REBOOTING = "Rebooting"
    DISCARDING = "Discarding"
    ALLOCATING = "Allocating"
    WAITING = "Waiting"
    CONNECTING = "Connecting"
    SENDING = "Sending"


class GeneralActionStatus(ListedEnum):
    ALLOCATED = "Allocated"
    CONNECTED = "Connected"


class Result(ListedEnum):
    FAILED = "Failed"
    SUCCESSFUL = "Successful"
    ABORTED = "Aborted"


OPERATIONS = (
    ExecutionPhase.list()
    + GeneralAction.list()
    + GeneralActionStatus.list()
    + ExecutionInfo.list()
    + Result.list()
)


# Start line
START_LINE = ".*Project content is packed for delivery.*"


def _match_date(date: str) -> bool:
    return re.match(r"\d{4}-\d{2}-\d{2}", date) is not None


def _match_time(time: str) -> bool:
    return re.match(r"\d{2}:\d{2}:\d{2}", time) is not None


def is_initial_line(line: str) -> bool:
    if not line:
        return False

    pattern = re.compile(START_LINE)
    parts = line.strip().split(" ")
    return (
        len(parts) > 2
        and _match_date(parts[0])
        and _match_time(parts[1])
        and re.match(pattern, line) is not None
    )


# Debug line starts with the operation
def is_operation(line: str, operation: Enum) -> bool:
    if not line:
        return False

    parts = line.strip().split(" ")
    return (
        len(parts) > 2
        and _match_date(parts[0])
        and _match_time(parts[1])
        and parts[2] == operation.value
    )


# Check if the line contains any operation
def is_any_operation(line: str) -> bool:
    if not line:
        return False

    parts = line.strip().split(" ")
    return (
        len(parts) > 2
        and _match_date(parts[0])
        and _match_time(parts[1])
        and parts[2] in OPERATIONS
    )


# Return the date in the line
def get_date(line: str) -> str:
    if not line:
        raise RuntimeError("Empty line provided")

    parts = line.strip().split(" ")
    if len(parts) <= 2 or not _match_date(parts[0]):
        raise ValueError("Incorrect format for line provided: {}".format(line))

    return parts[0]


# Return the time in the line
def get_time(line: str) -> str:
    if not line:
        raise RuntimeError("Empty line provided")

    parts = line.strip().split(" ")
    if len(parts) <= 2 or not _match_date(parts[0]) or not _match_time(parts[1]):
        raise ValueError("Incorrect format for line provided: {}".format(line))

    return parts[1]


# Return the operation in the line
def get_operation(line: str) -> str:
    if not line:
        raise RuntimeError("Empty line provided")

    parts = line.strip().split(" ")
    if len(parts) <= 2 or not _match_date(parts[0]) or not _match_time(parts[1]):
        raise ValueError("Incorrect format for line provided: {}".format(line))

    return parts[2]


# Return the information that comes after the operation
def get_operation_info(line: str) -> str:
    if not line:
        raise RuntimeError("Empty line provided")

    parts = line.strip().split(" ")
    if len(parts) <= 3 or not _match_date(parts[0]) or not _match_time(parts[1]):
        raise ValueError("Incorrect format for line provided: {}".format(line))

    return " ".join(parts[3:])


# Details are displayed after Error/Debug/Failed/Warning operations
def is_detail_start(line: str) -> bool:
    return (
        is_operation(line, ExecutionInfo.DEBUG)
        or is_operation(line, ExecutionInfo.ERROR)
        or is_operation(line, ExecutionInfo.WARNING)
        or is_operation(line, Result.FAILED)
    )


# Error/Debug/Failed output sometimes finish with either EOF error or a log file
# and no detail displayed
def is_detail(line: str) -> bool:
    return (
        is_operation(line, ExecutionInfo.DEBUG)
        or is_operation(line, ExecutionInfo.ERROR)
        or is_operation(line, ExecutionInfo.WARNING)
    ) and line.strip()[-1:] == ":"


# Error/Debug/Failed output finishes when a new other line starts
def is_detail_finished(line: str) -> bool:
    parts = line.strip().split(" ")
    return (
        len(parts) > 3
        and _match_date(parts[0])
        and _match_time(parts[1])
        and parts[2] in OPERATIONS
    )



#!/usr/bin/env python3
import sys
import dbus
import dbus.lowlevel

def run(pid, pid_start, uid, action_id):
    bus = dbus.SystemBus()
    obj = bus.get_object("org.freedesktop.PolicyKit1", "/org/freedesktop/PolicyKit1/Authority", False)
    polkit = dbus.Interface(obj, "org.freedesktop.PolicyKit1.Authority")
    subject = ("unix-process", {
        "pid": dbus.UInt32(pid, variant_level=1),
        "start-time": dbus.UInt64(pid_start, variant_level=1),
    })
    details = {"uid": uid}
    ret = polkit.CheckAuthorization(subject, action_id, details, dbus.UInt32(0), "")
    if len(ret) != 3:
        raise Exception("unexpected number of return values")

    if type(ret[0]) != dbus.types.Boolean:
        raise Exception("unexpected type of ret[0] %s, expected dbus.Boolean", type(ret[0]))
    if type(ret[1]) != dbus.types.Boolean:
        raise Exception("unexpected type of ret[1] %s, expected dbus.Boolean", type(ret[1]))
    if type(ret[2]) != dbus.types.Dictionary:
        raise Exception("unexpected type of ret[1] %s, expected dbus.Boolean", type(ret[2]))

    msg = dbus.lowlevel.Message()
    print(msg.guess_signature(ret), bool(ret[0]), bool(ret[1]), dict(ret[2]))


if __name__ == "__main__":
    pid = sys.argv[1]
    pid_start = sys.argv[2]
    uid = sys.argv[3]
    action_id = sys.argv[4]
    run(pid, pid_start, uid, action_id)



#!/usr/bin/env python3

import sys
from urllib.parse import urlparse, unquote

import dbus
import dbus.mainloop.glib
from gi.repository import GLib


class Client:
    def __init__(self, bus, main_loop):
        self._bus = bus
        self._main_loop = main_loop
        self._request_path = None
        self._response = None
        self._results = None
        bus.add_signal_receiver(
            self._response_cb,
            signal_name="Response",
            dbus_interface="org.freedesktop.portal.Request",
            bus_name="org.freedesktop.portal.Desktop",
            path_keyword="object_path",
        )
        self._portal = self._bus.get_object(
            "org.freedesktop.portal.Desktop",
            "/org/freedesktop/portal/desktop",
            introspect=False,
        )

    def _response_cb(self, response, results, object_path):
        if object_path == self._request_path:
            self._response = response
            self._results = results
            self._main_loop.quit()

    def run(self, args):
        command_name = args.pop(0).replace("-", "_")
        command = getattr(self, "cmd_{}".format(command_name), None)
        if command is None:
            sys.stderr.write("Unknown command: {}\n".format(command_name))
            sys.exit(1)
        return command(*args)

    def cmd_open_file(self):
        iface = dbus.Interface(self._portal, "org.freedesktop.portal.FileChooser")
        self._request_path = iface.OpenFile(
            "", "test open file", dbus.Dictionary(signature="sv")
        )
        self._main_loop.run()
        if self._response == 0:
            for uri in self._results["uris"]:
                parsed = urlparse(uri)
                assert parsed.scheme == "file"
                with open(unquote(parsed.path), "r") as fp:
                    sys.stdout.write(fp.read())
        elif self._response == 1:
            # user cancelled
            sys.stderr.write("request cancelled\n")
            return 1
        else:
            sys.stderr.write("request failed\n")
            return 1

    def cmd_save_file(self, content):
        iface = dbus.Interface(self._portal, "org.freedesktop.portal.FileChooser")
        self._request_path = iface.SaveFile(
            "", "test save file", dbus.Dictionary(signature="sv")
        )
        self._main_loop.run()
        if self._response == 0:
            uris = self._results["uris"]
            assert len(uris) == 1
            parsed = urlparse(uris[0])
            assert parsed.scheme == "file"
            with open(unquote(parsed.path), "w") as fp:
                fp.write(content)
        elif self._response == 1:
            # user cancelled
            sys.stderr.write("request cancelled\n")
            return 1
        else:
            sys.stderr.write("request failed\n")
            return 1

    def cmd_open_uri(self, uri):
        iface = dbus.Interface(self._portal, "org.freedesktop.portal.OpenURI")
        self._request_path = iface.OpenURI("", uri, dbus.Dictionary(signature="sv"))
        self._main_loop.run()
        if self._response == 0:
            pass
        elif self._response == 1:
            # user cancelled
            sys.stderr.write("request cancelled\n")
            return 1
        else:
            sys.stderr.write("request failed\n")
            return 1

    def cmd_launch_file(self, filename):
        iface = dbus.Interface(self._portal, "org.freedesktop.portal.OpenURI")
        with open(filename, "rb") as fp:
            self._request_path = iface.OpenFile(
                "", dbus.types.UnixFd(fp), dbus.Dictionary(signature="sv")
            )
        self._main_loop.run()
        if self._response == 0:
            pass
        elif self._response == 1:
            # user cancelled
            sys.stderr.write("request cancelled\n")
            return 1
        else:
            sys.stderr.write("request failed\n")
            return 1

    def cmd_screenshot(self):
        iface = dbus.Interface(self._portal, "org.freedesktop.portal.Screenshot")
        self._request_path = iface.Screenshot("", dbus.Dictionary(signature="sv"))
        self._main_loop.run()
        if self._response == 0:
            uri = self._results["uri"]
            parsed = urlparse(uri)
            assert parsed.scheme == "file"
            with open(unquote(parsed.path), "rb") as fp:
                sys.stdout.buffer.write(fp.read())
        elif self._response == 1:
            # user cancelled
            sys.stderr.write("request cancelled\n")
            return 1
        else:
            sys.stderr.write("request failed\n")
            return 1


def main(argv):
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    main_loop = GLib.MainLoop()
    bus = dbus.SessionBus()

    client = Client(bus, main_loop)
    return client.run(argv[1:])


if __name__ == "__main__":
    sys.exit(main(sys.argv))




import os
import sys
# To ensure the unit test can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import json
from io import StringIO
import unittest

import featextractor
from featextractor import CmdFeature, EndpointFeature, InterfaceFeature, TaskFeature, ChangeFeature, EnsureFeature 
from features import *
from state import State


def _get_stringio_from_loglines(loglines: list[dict]):
    lines=""
    for line in loglines:
        lines+=json.dumps(line) + "\n"
    return StringIO(lines)


class TestExtract(unittest.TestCase):
    

    def test_extract_cmd(self):
        cmd = "info snap"
        loglines = [{"msg":CmdFeature.msg, CmdLogLine.cmd:cmd}]
        logs = _get_stringio_from_loglines(loglines)
        d = featextractor.get_feature_dictionary(logs, ['cmd'], State({}))
        assert len(d) == 1
        assert 'cmds' in d and d['cmds'] == [Cmd(cmd=cmd)]

    def test_extract_endpoint(self):
        method = "GET"
        path = "/v2/snap/{info}"
        loglines = [{"msg":EndpointFeature.msg, EndpointLogLine.method:method, EndpointLogLine.path:path}]
        logs = _get_stringio_from_loglines(loglines)
        d = featextractor.get_feature_dictionary(logs, ['endpoint'], State({}))
        assert len(d) == 1
        assert 'endpoints' in d and d['endpoints'] == [Endpoint(method=method, path=path)]

    def test_extract_endpoint_with_action(self):
        method = "POST"
        path = "/v2/snap/{id}"
        action = "update"
        loglines = [{"msg":EndpointFeature.msg, EndpointLogLine.method:method, EndpointLogLine.path:path, EndpointLogLine.action:action}]
        logs = _get_stringio_from_loglines(loglines)
        d = featextractor.get_feature_dictionary(logs, ['endpoint'], State({}))
        assert len(d) == 1
        assert 'endpoints' in d and d['endpoints'] == [Endpoint(method=method, path=path, action=action)]

    def test_extract_interface(self):
        name = 'my-interface'
        plug = 'app'
        slot = 'snapd'
        loglines = [{"msg":InterfaceFeature.msg, InterfaceLogLine.interface:name, InterfaceLogLine.plug:plug, InterfaceLogLine.slot:slot}]
        logs = _get_stringio_from_loglines(loglines)
        d = featextractor.get_feature_dictionary(logs, ['interface'], State({}))
        assert len(d) == 1
        assert 'interfaces' in d and d['interfaces'] == [Interface(name=name, plug_snap_type=plug, slot_snap_type=slot)]

    def test_extract_ensure(self):
        mgr = 'my-manager'
        loglines = [
            {"msg":EnsureFeature.msg, EnsureLogLine.manager:mgr, EnsureLogLine.func:'1'},
            {"msg":EnsureFeature.msg, EnsureLogLine.manager:mgr, EnsureLogLine.func:'2'},
            {"msg":EnsureFeature.msg, EnsureLogLine.manager:mgr, EnsureLogLine.func:'3'},
                    ]
        logs = _get_stringio_from_loglines(loglines)
        d = featextractor.get_feature_dictionary(logs, ['ensure'], State({}))
        assert len(d) == 1
        assert 'ensures' in d and len(d['ensures']) == 3
        assert Ensure(manager=mgr, function='1') in d['ensures']
        assert Ensure(manager=mgr, function='2') in d['ensures']
        assert Ensure(manager=mgr, function='3') in d['ensures']
        
    def test_extract_task(self):
        loglines = [
            {"msg":TaskFeature.msg, TaskLogLine.id:"1", TaskLogLine.task_name:"kind1", TaskLogLine.status:"Doing"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"2", TaskLogLine.task_name:"kind2", TaskLogLine.status:"Doing"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"1", TaskLogLine.task_name:"kind1", TaskLogLine.status:"Done"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"3", TaskLogLine.task_name:"kind3", TaskLogLine.status:"Doing"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"3", TaskLogLine.task_name:"kind3", TaskLogLine.status:"Undoing"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"3", TaskLogLine.task_name:"kind3", TaskLogLine.status:"Undone"},
            {"msg":TaskFeature.msg, TaskLogLine.id:"2", TaskLogLine.task_name:"kind2", TaskLogLine.status:"Error"},
        ]
        logs = _get_stringio_from_loglines(loglines)
        state = {
            "tasks":{
                "1":{"kind":"test-task","data":{"snap-setup":{"type":"snapd"}}},
                "2":{"kind":"test-task","data":{"snap-type":"app"}},
                "3":{"kind":"test-task","data":{"snap-setup-task":"1"}}},
            }
        d = featextractor.get_feature_dictionary(logs, ['task'], State(state))
        assert len(d) == 1
        assert 'tasks' in d
        assert len(d['tasks']) == 3
        assert Task(kind="kind1", snap_types=["snapd"], last_status=Status.done.value) in d['tasks']
        assert Task(kind="kind2", snap_types=["app"], last_status=Status.error.value) in d['tasks']
        assert Task(kind="kind3", snap_types=["snapd"], last_status=Status.undone.value) in d['tasks']
        
    def test_extract_change(self):
        loglines = [
            {"msg":ChangeFeature.msg, ChangeLogLine.id:"1", ChangeLogLine.kind:"kind1"},
            {"msg":ChangeFeature.msg, ChangeLogLine.id:"2", ChangeLogLine.kind:"kind2"},
            {"msg":ChangeFeature.msg, ChangeLogLine.id:"3", ChangeLogLine.kind:"kind3"},
        ]
        logs = _get_stringio_from_loglines(loglines)
        state = {
            "changes":{
                "1":{"task-ids":["11"]},
                "2":{"task-ids":["12"]},
                "3":{},
            },
            "tasks":{
                "11":{"kind":"test-task","data":{"snap-setup":{"type":"snapd"}}},
                "12":{"kind":"test-task","data":{"snap-type":"app"}}
            }}
        d = featextractor.get_feature_dictionary(logs, ['change'], State(state))
        assert len(d) == 1
        assert 'changes' in d and len(d['changes']) == 3
        assert Change(kind="kind1", snap_types=["snapd"]) in d['changes']
        assert Change(kind="kind2", snap_types=["app"]) in d['changes']
        assert Change(kind="kind3", snap_types=[]) in d['changes']

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import argparse
import datetime
import json
import os
from pymongo import MongoClient, InsertOne

HOST = 'MONGO_HOST'
PORT = 'MONGO_PORT'
USER = 'MONGO_USER'
PASSWORD = 'MONGO_PASSWORD'


def upload_documents(folder, verbose):
    if HOST not in os.environ.keys():
        raise RuntimeError(
            "the {} environment variable must be set and contain the host data".format(HOST))
    if PORT not in os.environ.keys():
        raise RuntimeError(
            "the {} environment variable must be set and contain the port data".format(PORT))
    if USER not in os.environ.keys():
        raise RuntimeError(
            "the {} environment variable must be set and contain the username".format(USER))
    if PASSWORD not in os.environ.keys():
        raise RuntimeError(
            "the {} environment variable must be set and contain the password".format(PASSWORD))
    if not os.environ[PORT].isdigit():
        raise RuntimeError(
            "the {} environment variable must contain a valid port number".format(PORT))

    with MongoClient(host=os.environ[HOST], port=int(os.environ[PORT]), username=os.environ[USER], password=os.environ[PASSWORD]) as client:
        db = client.snapd
        collection = db.features

        requesting = []
        timestamp = datetime.datetime.now(datetime.timezone.utc)
        for file in os.listdir(folder):
            if file.endswith(".json"):
                with open(os.path.join(folder, file), 'r', encoding='utf-8') as f:
                    j = json.load(f)
                    j['timestamp'] = timestamp
                    if file == 'all-features.json':
                        j['all_features'] = True
                    requesting.append(InsertOne(j))

        result = collection.bulk_write(requesting)
        if verbose:
            print("inserted {} new documents".format(result.inserted_count))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="mongodb document uploader. It assumes the following environment variables are set: {}, {}, {}, and {}".format(HOST, PORT, USER, PASSWORD))
    parser.add_argument(
        '--dir', help='directory containing json files', required=True, type=str)
    parser.add_argument(
        '--verbose', help='print upload statement', action='store_true')
    args = parser.parse_args()

    if not os.path.isdir(args.dir):
        raise RuntimeError(
            "the indicated directory {} does not exist.".format(args.dir))

    upload_documents(args.dir, args.verbose)



import json
import sys

doc = json.load(sys.stdin)

slot_to_token = {}
for k, v in doc.get('tokens', {}).items():
    for slot in v.get('keyslots', []):
        slot_to_token[slot] = k

orphans = []
for k, v in doc.get('keyslots', {}).items():
    if k not in slot_to_token:
        orphans.append(k)

if len(orphans) > 0:
    formatted = ', '.join(orphans)
    print(f'orphan key slots: {formatted}', file=sys.stderr)
    sys.exit(1)
else:
    sys.exit(0)



# Copyright (C) 2022 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import argparse

# The vendor certificate table in created by cert.S in the shim project.
# See:
# https://github.com/rhboot/shim/blob/505cdb678b319fcf9a7fdee77c0f091b4147cbe5/cert.S

def main():
    parser = argparse.ArgumentParser(description='Generate a shim vendor certificate section')
    parser.add_argument('OUTPUT', help='Output file')
    parser.add_argument('CERT', help='Vendor certificate')
    parser.add_argument('DBX', nargs='?', help='Vendor dbx')
    args = parser.parse_args()
    with open(args.CERT, 'rb') as f:
        cert = f.read()
    if args.DBX is not None:
        with open(args.DBX, 'rb') as f:
            dbx = f.read()
    else:
        dbx = b''

    header_format = 'IIII'
    header_size = struct.calcsize(header_format)
    header = struct.pack(header_format, len(cert), len(dbx), header_size, header_size+len(cert))
    with open(args.OUTPUT, 'wb') as f:
        f.write(header)
        f.write(cert)
        f.write(dbx)

if __name__ == '__main__':
    main()
